from tests.unit import UnitTestBase
from express.properties.scalar.total_energy import TotalEnergy

TOTAL_ENERGY = {
    "units": "eV",
    "name": "total_energy",
    "value": 1
}


class TotalEnergyTest(UnitTestBase):
    def setUp(self):
        super(TotalEnergyTest, self).setUp()

    def tearDown(self):
        super(TotalEnergyTest, self).setUp()

    def test_total_energy(self):
        parser = self.get_mocked_parser("total_energy", 1)
        property_ = TotalEnergy("total_energy", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), TOTAL_ENERGY)





import os
import xml.etree.ElementTree as ET


class BaseXMLParser(object):
    """
    Base XML parser class.
    """

    def __init__(self, xml_file_path):
        self.xml_path = xml_file_path
        self.root = None
        self.xml_dir_name = None
        if self.xml_path and os.path.exists(self.xml_path):
            try:
                self.xml_dir_name = os.path.dirname(self.xml_path)
                self.root = ET.parse(self.xml_path).getroot()
            except:
                # safely ignore broken xml file
                pass



import os
import re
import io
import numpy as np

from express.parsers.utils import find_file
from express.parsers.settings import Constant
from express.parsers.apps.espresso import settings
from express.parsers.formats.txt import BaseTXTParser

ORBITS = {
    's': [''],
    'p': ['z', 'x', 'y'],
    'd': ['z2', 'zx', 'zy', 'x2-y2', 'xy']
}


class EspressoTXTParser(BaseTXTParser):
    """
    Espresso text parser class.
    """

    def __init__(self, work_dir):
        super(EspressoTXTParser, self).__init__(work_dir)

    def total_energy(self, text):
        """
        Extracts total energy.

        Args:
            text (str): text to extract data from.

        Returns:
             float
        """
        return Constant.RYDBERG * self._general_output_parser(text, **settings.REGEX["total_energy"])

    def dos(self):
        """
        Extracts density of states. It reads 'pdos_tot' file to extract energy levels and total DOS values for each
        energy level. Then it reads partial DOS files created by QE with this format: `__prefix__.pdos_atm#1(
        C)_wfc#1(s)` in job working directory. DOS value for each atom with the same element and orbit number will be
        added together and packed in a dictionary. The result containing energy levels, total DOS and partial DOS for
        each element will be returned.

        Returns:
            dict

        Example:
            {
                'energy': [-1.0, 0, 1.0],
                'total': [0.013, 0.124, 0.923],
                'partial': [
                    {
                        'element': 'C',
                        'electronicState': '1s',
                        'value': [0.00015, 0.000187, 0.000232, 0.000287, 0.000355, 0.000437]
                    },
                    {
                        'element': 'C',
                        'electronicState': '3px',
                        'value': [6.87e-06, 8.5e-06, 1.0e-05, 1.3e-05, 1.63e-05, 2.01e-05]
                    },
                ]
            }
        """
        dos_tot_file = find_file(settings.PDOS_TOT_FILE, self.work_dir)
        energy_levels, total_dos = self._total_dos(dos_tot_file)
        partial_dos_values, partial_dos_infos = self._partial_dos(len(energy_levels))
        return {
            'energy': energy_levels.tolist(),
            'total': total_dos.tolist(),
            'partial': partial_dos_values,
            'partial_info': partial_dos_infos
        }

    def _total_dos(self, dos_tot_file):
        """
        Parses total DOS from the given file. It reads the file to get energy levels and total DOS values for each
        energy level by extracting the first two columns.

        Example file:
            #
             E (eV)  dos(E)    pdos(E)
            -25.226  0.512E-03  0.504E-03
            -25.216  0.637E-03  0.628E-03
            -25.206  0.791E-03  0.779E-03

        Args:
            dos_tot_file (str): path to the pdos_tot file.

        Returns:
            tuple: energy levels and DOS total values.
                   Example:
                       ([-25.226, -25.216, -25.206], [0.512E-03, 0.637E-03, 0.791E-03])
        """
        if os.path.isfile(dos_tot_file):
            trimmed_dos_file = io.StringIO(self._trim_dos_file(dos_tot_file))
            dos_tot = np.genfromtxt(trimmed_dos_file, dtype=np.float32, usecols=(0, 1))
            energy_levels = dos_tot[:, 0]
            dos_tot = dos_tot[:, 1]
            return energy_levels, dos_tot

    def _partial_dos(self, num_levels):
        """
        Parses partial DOS for each element with its orbit value. it reads partial DOS files created by QE with this
        format: `__prefix__.pdos_atm#1(C)_wfc#1(s)` in job working directory. DOS value for each atom with the same
        element and orbit number will be added together and packed in a dictionary.

        Args:
            num_levels (int): number of energy levels.

        Returns:
            dict: a dictionary containing partial DOS values for each element.
                Example:
                    [
                        {
                            'element': 'C',
                            'electronicState': '1s',
                            'value': [0.00015, 0.000187, 0.000232, 0.000287, 0.000355, 0.000437]
                        },
                        {
                            'element': 'C',
                            'electronicState': '3px',
                            'value': [6.87e-06, 8.5e-06, 1.0e-05, 1.3e-05, 1.63e-05, 2.01e-05]
                        }
                    ]
        """
        pdos = {}
        # Because os.listdir() has an undefined order specified, we'll sort the file list in order to have a
        # reproducible result.  The sort order will be the normal alphanumeric comparison.
        # For example:
        # >>> x = ['a', 'B', 'c', 'D']
        # >>> sorted(x)
        # ['B', 'D', 'a', 'c']
        for file_name in sorted(os.listdir(self.work_dir)):
            file_path = os.path.join(self.work_dir, file_name)
            match = re.compile(settings.REGEX['pdos_file']['regex']).match(file_name)
            if match:
                atm_pdos = self._extract_partial_dos(file_path, len(ORBITS[match.group('orbit_symbol')]))
                atm_pdos = atm_pdos.T if atm_pdos.shape[0] > 1 else atm_pdos
                for idx, orbit_pdos in enumerate(atm_pdos):
                    orbit_idx = ORBITS[match.group('orbit_symbol')][idx] if match.group('orbit_symbol') != 's' else ''
                    pdos_id = "{0}_{1}{2}{3}".format(match.group('atom_name'), match.group('orbit_num'),
                                                     match.group('orbit_symbol'), orbit_idx)  # e.g. C_1s, C_2px, C_2dz2
                    if not pdos_id in pdos.keys():
                        pdos[pdos_id] = np.zeros(num_levels)
                    pdos[pdos_id] += orbit_pdos

        pdos_values = [pdos[item].tolist() for item in pdos]
        pdos_infos = [{'element': item.split('_')[0], 'electronicState': item.split('_')[1]} for item in pdos]
        return pdos_values, pdos_infos

    def _extract_partial_dos(self, pdos_file, orbit_num):
        """
        Parses partial DOS values from a given file.

        Args:
            pdos_file (str): path to pdos file.

        Returns:
            numpy.ndarray
        """
        if os.path.isfile(pdos_file):
            trimmed_pdos_file = io.StringIO(self._trim_dos_file(pdos_file))
            target_columns = range(2, 2 + orbit_num)
            columns = np.genfromtxt(trimmed_pdos_file, dtype=np.float32, usecols=range(2, 2 + orbit_num))
            return columns if len(target_columns) > 1 else np.array([columns])

    def _trim_dos_file(self, dos_file):
        """
        Trims a given dos file and returns the data. Only lines which start with a scientific number are interested.

        Args:
            dos_file (str): path to the dos file.

        Returns:
            (str): trimmed out text containing actual dos values.
                Example Input:
                    *******  0.000E+00  0.000E+00
                    *******  0.000E+00  0.000E+00
                    -99.989  0.000E+00  0.000E+00
                     99.939  0.000E+00  0.000E+00
                      9.889  0.000E+00  0.000E+00
                    *******  0.000E+00  0.000E+00
                    *******  0.000E+00  0.000E+00
                Example output:
                    -99.989  0.000E+00  0.000E+00
                    -99.939  0.000E+00  0.000E+00
                    -99.889  0.000E+00  0.000E+00
        """
        with open(dos_file) as f:
            return '\n'.join(re.findall('^ *[-+]?\d*\.\d+(?:[eE][-+]?\d+)?.*$', f.read(), re.MULTILINE))

    def convergence_electronic(self, text):
        """
        Extracts convergence electronic.

        Args:
            text (str): text to extract data from.

        Returns:
             list[float]
        """
        data = self._general_output_parser(text, **settings.REGEX["convergence_electronic"])
        # The next 3 lines are necessary to have realtime data
        ionic_data = [_["electronic"]["data"] for _ in self.convergence_ionic(text)]
        last_step_data = data[sum([len(_) for _ in ionic_data]): len(data)]
        if last_step_data: ionic_data.append(last_step_data)
        return [(np.array(_) * Constant.RYDBERG).tolist() for _ in ionic_data]

    def convergence_ionic(self, text):
        """
        Extracts convergence ionic.

        Args:
            text (str): text to extract data from.

        Returns:
             list[dict]
        """
        data = []
        blocks = re.findall(settings.REGEX["convergence_ionic_blocks"]["regex"], text, re.DOTALL | re.MULTILINE)
        for idx, block in enumerate(blocks):
            energies = self._general_output_parser(block, **settings.REGEX["convergence_ionic_energies"])
            energies = (np.array(energies) * Constant.RYDBERG).tolist()
            data.append({
                "energy": energies[-1],
                "electronic": {
                    "units": "eV",
                    "data": self._general_output_parser(block, **settings.REGEX["convergence_electronic"])
                },
            })

        if not data: return []

        # last structure is used for the next ionic step, hence [:max(0, len(data) - 1)]
        lattice_convergence = self._lattice_convergence(text)[:max(0, len(data) - 1)]
        basis_convergence = self._basis_convergence(text)[:max(0, len(data) - 1)]
        for idx, structure in enumerate(zip(lattice_convergence, basis_convergence)):
            structure[1]["units"] = "angstrom"
            lattice_matrix = np.array([structure[0]["vectors"][key] for key in ["a", "b", "c"]]).reshape((3, 3))
            for coordinate in structure[1]["coordinates"]:
                coordinate["value"] = np.dot(coordinate["value"], lattice_matrix).tolist()
            data[idx + 1].update({
                'structure': {
                    'lattice': structure[0],
                    'basis': structure[1]
                }
            })

        # inject initial structure
        data[0].update({
            "structure": {
                "basis": self.initial_basis(text),
                "lattice": self.initial_lattice_vectors(text)
            }
        })

        return data

    def initial_lattice_vectors(self, text):
        """
        Extracts initial lattice from a given text.

        Note: The initial lattice is in alat format and hence it needs to be converted to angstrom.

        The text looks like the following:

            crystal axes: (cart. coord. in units of alat)
               a(1) = (   0.866025   0.000000   0.500000 )
               a(2) = (   0.288675   0.816497   0.500000 )
               a(3) = (   0.000000   0.000000   1.000000 )

        Args:
            text (str): text to extract data from.

        Returns:
            dict

        Example:
            {
                'vectors': {
                    'a': [-0.561154473, -0.000000000, 0.561154473],
                    'b': [-0.000000000, 0.561154473, 0.561154473],
                    'c': [-0.561154473, 0.561154473, 0.000000000],
                    'alat': 1
                }
            }
        """
        alat = self._get_alat(text)
        lattice_in_alat_units = self._extract_lattice(text, regex="lattice_alat")
        for key in ["a", "b", "c"]:
            lattice_in_alat_units["vectors"][key] = [e * alat * Constant.BOHR for e in lattice_in_alat_units["vectors"][key]]
        return lattice_in_alat_units

    def _extract_basis(self, text, number_of_atoms):
        """
        Extracts the basis from the given text.

        Note: no units conversion is done in here.

        """
        basis = {"units": "angstrom", "elements": [], "coordinates": []}
        matches = self._general_output_parser(text, **settings.REGEX["basis_alat"](number_of_atoms))
        for idx, match in enumerate(matches):
            basis["elements"].append({"id": idx, "value": match[0]})
            coordinate = [float(match[1]), float(match[2]), float(match[3])]
            basis["coordinates"].append({"id": idx, "value": coordinate})
        return basis

    def _get_alat(self, text):
        return self._general_output_parser(text, **settings.REGEX["lattice_parameter_alat"])[0]

    def _number_of_atoms(self, text):
        return self._general_output_parser(text, **settings.REGEX["number_of_atoms"])[0]

    def initial_basis(self, text):
        """
        Extracts initial basis from a given text.

        Units: angstrom

        The text looks like the following:

             site n.     atom                  positions (alat units)
                 1           Si  tau(   1) = (   0.0000000   0.0000000   0.0000000  )
                 2           Si  tau(   2) = (   0.2886752   0.2041241   0.5000000  )

        Args:
            text (str): text to extract data from.

        Returns:
            dict

         Example:
            {
                'units': 'angstrom',
                'elements': [{'id': 1, 'value': 'Si'}, {'id': 2, 'value': 'Si'}],
                'coordinates': [{'id': 1, 'value': [0.0, 0.0, 0.0]}, {'id': 2, 'value': [2.1095228, 1.49165, 3.6538]}]
             }
        """
        alat = self._get_alat(text)
        number_of_atoms = self._number_of_atoms(text)
        basis_in_alat_units = self._extract_basis(text[text.find("positions (alat units)"):], number_of_atoms)
        for coordinate in basis_in_alat_units["coordinates"]:
            coordinate["value"] = [x * alat * Constant.BOHR for x in coordinate["value"]]
        return basis_in_alat_units

    def _lattice_convergence(self, text):
        """
        Extracts lattice convergence values in angstrom units computed in each BFGS step.

        Args:
            text (str): text to extract data from.

        Returns:
            list

        Example:
            [
                {
                    'vectors': {
                        'a': [-0.561154473, -0.000000000, 0.561154473],
                        'b': [-0.000000000, 0.561154473, 0.561154473],
                        'c': [-0.561154473, 0.561154473, 0.000000000],
                        'alat': 1
                    }
                 },
                ...
                {
                    'vectors': {
                        'a': [-0.561154473, -0.000000000, 0.561154473],
                        'b': [-0.000000000, 0.561154473, 0.561154473],
                        'c': [-0.561154473, 0.561154473, 0.000000000],
                        'alat': 1
                    }
                 }
             ]
        """
        return self._extract_data_from_bfgs_blocks(text, self._extract_lattice)

    def _extract_data_from_bfgs_blocks(self, text, func):
        """
        Extracts data from BFGS blocks using the provided function.

        Args:
            text (str): text to extract data from.
            func (func): function object to be applied on the content block.

        Returns:
            list: list of information extracted using the provided function
        """
        results = []
        bfgs_block_pattern = re.compile(settings.REGEX["bfgs_block"]["regex"], re.DOTALL)
        bfgs_blocks = bfgs_block_pattern.findall(text)
        for block in bfgs_blocks:
            results.append(func(block))
        return results

    def _extract_lattice(self, text, regex="lattice"):
        """
        Extracts lattice.

        Note: no units conversion is done in here.

        Args:
            text (str): text to extract data from.

        Returns:
            dict

        Example:
            {
                'vectors': {
                    'a': [-0.561154473, -0.000000000, 0.561154473],
                    'b': [-0.000000000, 0.561154473, 0.561154473],
                    'c': [-0.561154473, 0.561154473, 0.000000000],
                    'alat': 1
                }
            }
        """
        match = re.search(settings.REGEX[regex]["regex"], text)
        if match:
            lattice = [float(_) for _ in match.groups(1)]
            return {
                'vectors': {
                    'a': lattice[0:3],
                    'b': lattice[3:6],
                    'c': lattice[6:9],
                    'alat': 1
                }
            }

    def _basis_convergence(self, text):
        """
        Extracts convergence bases computed in each BFGS step.

        Args:
            text (str): text to extract data from.

        Returns:
            list

        Example:
            [
                {
                    'units': 'crystal',
                    'elements': [{'id': 1, 'value': 'Si'}, {'id': 2, 'value': 'Si'}],
                    'coordinates': [{'id': 1, 'value': [0.0, 0.0, 0.0]}, {'id': 2, 'value': [0.0, 0.0, 0.0]}]
                 },
                ...
                {
                    'units': 'crystal',
                    'elements': [{'id': 1, 'value': 'Si'}, {'id': 2, 'value': 'Si'}],
                    'coordinates': [{'id': 1, 'value': [0.0, 0.0, 0.0]}, {'id': 2, 'value': [0.0, 0.0, 0.0]}]
                 }
            ]
        """
        return self._extract_data_from_bfgs_blocks(text, self._extract_basis_from_bfgs_blocks)

    def _extract_basis_from_bfgs_blocks(self, text):
        """
        Extracts basis data in crystal units.

        Args:
            text (str): text to extract data from.

        Returns:
            dict

        Example:
            {
                'units': 'crystal',
                'elements': [{'id': 1, 'value': 'Si'}, {'id': 2, 'value': 'Si'}],
                'coordinates': [{'id': 1, 'value': [0.0, 0.0, 0.0]}, {'id': 2, 'value': [0.0, 0.0, 0.0]}]
             }
        """
        basis = {
            "units": "crystal",
            "elements": [],
            "coordinates": []
        }
        matches = re.findall(settings.REGEX["ion_position"]["regex"], text)
        if matches:
            for idx, match in enumerate(matches):
                basis["elements"].append({
                    "id": idx,
                    "value": match[0]
                })
                basis["coordinates"].append({
                    "id": idx,
                    "value": [float(match[1]), float(match[2]), float(match[3])]
                })

            return basis

    def stress_tensor(self, text):
        """
        Extracts stress tensor.

        Args:
            text (str): text to extract data from.

        Returns:
            list
        """
        return self._general_output_parser(text, **settings.REGEX["stress_tensor"])

    def pressure(self, text):
        """
        Extracts pressure.

        Args:
            text (str): text to extract data from.

        Returns:
            float
        """
        return self._general_output_parser(text, **settings.REGEX["pressure"])

    def total_force(self, text):
        """
        Extracts total force.

        Returns:
            float
        """
        return self._general_output_parser(text, **settings.REGEX["total_force"]) * Constant.ry_bohr_to_eV_A

    def atomic_forces(self, text):
        """
        Extracts atomic forces.

        Args:
            text (str): text to extract data from.

        Returns:
            list
        """
        forces = self._general_output_parser(text, **settings.REGEX['forces_on_atoms'])
        return (np.array(forces) * Constant.ry_bohr_to_eV_A).tolist()

    def total_energy_contributions(self, text):
        """
        Extracts total energy contributions.

        Args:
            text (str): text to extract data from.

        Returns:
            dict
        """
        energy_contributions = {}
        for contribution in settings.TOTAL_ENERGY_CONTRIBUTIONS:
            value = self._general_output_parser(text, **settings.TOTAL_ENERGY_CONTRIBUTIONS[contribution])
            if value is not None:
                energy_contributions.update({contribution: {
                    'name': contribution,
                    'value': value * Constant.RYDBERG
                }})
        return energy_contributions

    def zero_point_energy(self, text):
        """
        Extracts zero point energy.

        Returns:
             float
        """
        data = self._general_output_parser(text, **settings.REGEX['zero_point_energy'])
        if len(data): return (sum(data) / 2) * Constant.cm_inv_to_ev

    def phonon_dos(self):
        """
        Extract vibrational frequencies and total DOS.

        Returns:
            dict

        Example:
            {
                'frequency': [-1.2588E-05, 9.9999E-01, 2.0000E+00, 3.0000E+00, ....]
                'total': [0.0000E+00, 2.5386E-07, 1.0154E-06, 2.2847E-06, ....]
            }
        """
        phonon_dos_tot_file = find_file(settings.PHONON_DOS_FILE, self.work_dir)
        frequencies, total_phonon_dos = self._total_dos(phonon_dos_tot_file)
        return {
            'frequency': frequencies.tolist(),
            'total': total_phonon_dos.tolist()
        }

    def phonon_dispersions(self):
        """
        Extract vibrational frequencies at qpoints along the high symmetry points in Brillouin zone.

        Returns:
            dict

        Example:
            {
                'qpoints': [[0.00, 0.00, 0.00],[0.00, 0.00, 0.01],....],
                'frequencies': [['-0.0000', '-0.0000', '-0.0000', '574.0778', '574.0778', '574.2923'],
                ['29.3716', '30.0630', '70.4699', '568.0790', '569.7664', '569.9710'], ....]
            }
        """
        modes_file = find_file(settings.PHONON_MODES_FILE, self.work_dir)
        qpoints, frequencies = self.phonon_frequencies(modes_file)
        return {
            'qpoints': qpoints.tolist(),
            'frequencies': frequencies.tolist()
        }

    def phonon_frequencies(self, modes_file):
        """
        Extracts qpoints along the paths in Brillouin zone and the phonon frequencies at each qpoint.

        Example file:
            #
             q =       0.0000      0.0000      0.0000
             **************************************************************************
             freq (    1) =      -0.004399 [THz] =      -0.146739 [cm-1]
             ( -0.366864   0.000000     0.211786   0.000000     0.265747   0.000000   )
             ( -0.367203   0.000000     0.211982   0.000000     0.264868   0.000000   )
             ( -0.367036   0.000000     0.211885   0.000000     0.265493   0.000000   )
             ( -0.367206   0.000000     0.211984   0.000000     0.264823   0.000000   )
             freq (    2) =      -0.004372 [THz] =      -0.145845 [cm-1]
             ....
             ....
             **************************************************************************
             q =       0.0000      0.0000      0.1000
             **************************************************************************
             freq (    1) =      -0.004399 [THz] =      -0.146739 [cm-1]
             ( -0.366864   0.000000     0.211786   0.000000     0.265747   0.000000   )
             ( -0.367203   0.000000     0.211982   0.000000     0.264868   0.000000   )
             ( -0.367036   0.000000     0.211885   0.000000     0.265493   0.000000   )
             ( -0.367206   0.000000     0.211984   0.000000     0.264823   0.000000   )
             freq (    2) =      -0.004372 [THz] =      -0.145845 [cm-1]
             ....
             ....
             **************************************************************************

        Args:
            modes_file (str): path to modes file (normal_modes.out).

        Returns:
            tuple

        Example:
            ([
                [0.0000, 0.0000, 0.0000],
                [0.000, 0.0000, 0.1000]
            ],
            [
                [5.7429E+02, 5.7429E+02],
                [5.69970E+02, 5.69970E+02],
                .....,
                [4.5469E+02, 4.5469E+02]
            ])
        """
        with open(modes_file, 'r') as f:
            text = f.read()
        qpoints = np.array(re.compile(settings.REGEX["qpoints"]["regex"]).findall(text), dtype=np.float)
        frequencies = np.array(re.compile(settings.REGEX["phonon_frequencies"]["regex"]).findall(text), dtype=np.float)
        frequencies = np.transpose(frequencies.reshape(qpoints.shape[0], frequencies.shape[0] // qpoints.shape[0]))
        return qpoints, frequencies

    def reaction_coordinates(self, text):
        """
        Extracts reaction coordinates from the first column of NEB dat file.

        Example input:
            0.0000000000      0.0000000000      0.1145416373
            0.1944842569      0.0341286086      0.0886712471
            0.3625057164      0.1308566687      0.0995133799
            0.4999390812      0.2030519699      0.0010169552
            0.6383139128      0.1302022095      0.0385626296
            0.8046562342      0.0345676488      0.0062796586
            1.0000000000      0.0000000063      0.1146893883


        Returns:
            list[float]
        """
        return self._general_output_parser(text, **settings.REGEX["reaction_coordinates"])

    def reaction_energies(self, text):
        """
        Extracts reaction energies from the second column of NEB dat file.

        Example input:
            0.0000000000      0.0000000000      0.1145416373
            0.1944842569      0.0341286086      0.0886712471
            0.3625057164      0.1308566687      0.0995133799
            0.4999390812      0.2030519699      0.0010169552
            0.6383139128      0.1302022095      0.0385626296
            0.8046562342      0.0345676488      0.0062796586
            1.0000000000      0.0000000063      0.1146893883


        Returns:
            list[float]
        """
        return self._general_output_parser(text, **settings.REGEX["reaction_energies"])

    def potential_profile(self, text):
        """
        Extracts potential (hartree, local, hartree+local) along z.

        Example input:
            #z (A)  Tot chg (e/A)  Avg v_hartree (eV)  Avg v_local (eV)  Avg v_hart+v_loc (eV)
             -4.89        2.3697          -6.5847438         6.4872255        -0.0975183
             -4.78        2.1422          -7.0900648         8.2828137         1.1927490
             -4.67        2.0006          -7.5601238        10.1322914         2.5721676
             -4.56        1.8954          -7.9743444        11.9417738         3.9674294
             -4.44        1.7923          -8.3174980        13.6185904         5.3010924
             -4.33        1.6879          -8.5750414        15.0903496         6.5153082
             -4.22        1.5891          -8.7323531        16.2665057         7.5341527
             -4.11        1.5036          -8.7756094        17.0759068         8.3002974
             -4.00        1.4383          -8.6928512        17.5243394         8.8314882
             -3.89        1.3984          -8.4749404        17.6353196         9.1603792

        Returns:
            list[list[float]]
        """
        data = self._general_output_parser(text, **settings.REGEX["potential_profile"])
        return [[e[i] for e in data] for i in range(4)]

    def charge_density_profile(self, text):
        """
        Extracts total charge density along z.

        Example input:
            #z (A)  Tot chg (e/A)  Avg v_hartree (eV)  Avg v_local (eV)  Avg v_hart+v_loc (eV)
             -4.89        2.3697          -6.5847438         6.4872255        -0.0975183
             -4.78        2.1422          -7.0900648         8.2828137         1.1927490
             -4.67        2.0006          -7.5601238        10.1322914         2.5721676
             -4.56        1.8954          -7.9743444        11.9417738         3.9674294
             -4.44        1.7923          -8.3174980        13.6185904         5.3010924
             -4.33        1.6879          -8.5750414        15.0903496         6.5153082
             -4.22        1.5891          -8.7323531        16.2665057         7.5341527
             -4.11        1.5036          -8.7756094        17.0759068         8.3002974
             -4.00        1.4383          -8.6928512        17.5243394         8.8314882
             -3.89        1.3984          -8.4749404        17.6353196         9.1603792

        Returns:
            list[list[float]]
        """
        data = self._general_output_parser(text, **settings.REGEX["charge_density_profile"])
        return [[e[i] for e in data] for i in range(2)]

    def eigenvalues_at_kpoints_from_sternheimer_gw_stdout(self, text, inverse_reciprocal_lattice_vectors):
        """
        Extracts eigenvalues for all kpoints from Sternheimer GW stdout file.

        Example input:

            LDA eigenval (eV)   -5.60    6.25    6.25    6.25    8.69    8.69    8.69    9.37

            GWKpoint cart :   0.0000   0.0000   0.0000

            GWKpoint cryst:   0.0000   0.0000   0.0000
            GW qp energy (eV)   -7.08    5.34    5.34    5.34   10.15   10.15   10.15   10.82
            Vxc expt val (eV)  -10.09  -10.20  -10.20  -10.20   -9.85   -9.85   -9.85  -10.34
            Sigma_ex val (eV)  -16.04  -15.82  -15.82  -15.82   -3.54   -3.54   -3.54   -4.03
            QP renorm            1.00    1.00    1.00    1.00    1.00    1.00    1.00    1.00

        Returns:
            list[]
        """
        kpoints = self._general_output_parser(text, **settings.REGEX["sternheimer_gw_kpoint"])
        eigenvalues = self._general_output_parser(text, **settings.REGEX["sternheimer_gw_eigenvalues"])
        eigenvalues = [[float(x) for x in re.sub(' +', ' ', e).strip(" ").split(" ")] for e in eigenvalues]
        return [
            {
                'kpoint': np.dot(point, inverse_reciprocal_lattice_vectors).tolist(),
                'weight': 1.0 / len(kpoints),  # uniformly set the weights as they are not extractable.
                'eigenvalues': [
                    {
                        'energies': eigenvalues[index],
                        'occupations': [],  # set occupations empty as they are not extractable.
                        'spin': 0.5  # spin-polarized calculation is not supported yet, hence 0.5
                    }
                ]
            } for index, point in enumerate(kpoints)]

    def final_basis(self, text):
        """
        Extracts final basis in angstrom units.
        """
        atomic_position_last_index = text.rfind("ATOMIC_POSITIONS (crystal)")
        if atomic_position_last_index < 0: return self.initial_basis(text)
        number_of_atoms = self._number_of_atoms(text)
        basis = self._extract_basis(text[atomic_position_last_index:], number_of_atoms)

        # final basis is in crystal units, hence it needs to be converted into angstrom.
        final_lattice_vectors = self.final_lattice_vectors(text)
        lattice_matrix = np.array([final_lattice_vectors["vectors"][key] for key in ["a", "b", "c"]]).reshape((3, 3))
        for coordinate in basis["coordinates"]:
            coordinate["value"] = np.dot(coordinate["value"], lattice_matrix).tolist()

        return basis

    def final_lattice_vectors(self, text):
        """
        Extracts final lattice in angstrom units.
        """
        cell_parameters_last_index = text.rfind("CELL_PARAMETERS (angstrom)")
        if cell_parameters_last_index < 0: return self.initial_lattice_vectors(text)
        return self._extract_lattice(text[cell_parameters_last_index:])



from express.properties.non_scalar.two_dimensional_plot import TwoDimensionalPlotProperty


class PhononDispersions(TwoDimensionalPlotProperty):
    """
    Describes the frequencies and shapes of possible atomic oscillations inside a material.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(PhononDispersions, self).__init__(name, parser, *args, **kwargs)
        phonon_dispersions = self.parser.phonon_dispersions()
        self.xDataArray = phonon_dispersions['qpoints']
        self.yDataSeries = phonon_dispersions['frequencies']



from express.properties.non_scalar import NonScalarProperty


class InchiKey(NonScalarProperty):
    """
    Inchi key property class.
    """
    def __init__(self, name, parser, *args, **kwargs):
        super(InchiKey, self).__init__(name, parser, *args, **kwargs)
        self.name = name
        self.value = self.parser.get_inchi_key()

    def _serialize(self):
        return self.value



def eigenvalues(eigenvalues_at_kpoints, kpoint_index=0, spin_index=0):
    """
    Returns eigenvalues for a given kpoint and spin.

    Args:
        eigenvalues_at_kpoints (list): a list of eigenvalues for all kpoints.
        kpoint_index (int): kpoint index.
        spin_index (int): spin index.

    Returns:
         ndarray
    """
    spin_map = {0: 0.5, 1: -0.5}
    for eigenvalue in eigenvalues_at_kpoints[kpoint_index]["eigenvalues"]:
        if eigenvalue['spin'] == spin_map[spin_index]:
            return eigenvalue['energies']


def to_array_with_ids(array):
    """
    Converts a given array to an array of objects with id.

    Args:
        array (list): array to convert

    Returns:
        list
    """
    return [{"id": index + 1, "value": value} for index, value in enumerate(array)]



import os
import numpy as np

from express.parsers import BaseParser
from express.parsers.apps.espresso import settings
from express.parsers.mixins.ionic import IonicDataMixin
from express.parsers.mixins.reciprocal import ReciprocalDataMixin
from express.parsers.mixins.electronic import ElectronicDataMixin
from express.parsers.utils import find_file, lattice_basis_to_poscar
from express.parsers.apps.espresso.settings import NEB_PATH_FILE_SUFFIX
from express.parsers.apps.espresso.formats.txt import EspressoTXTParser
from express.parsers.apps.espresso.formats.xml import EspressoXMLParser


class EspressoParser(BaseParser, IonicDataMixin, ElectronicDataMixin, ReciprocalDataMixin):
    """
    Espresso parser class.
    """

    def __init__(self, *args, **kwargs):
        super(EspressoParser, self).__init__(*args, **kwargs)
        self.work_dir = self.kwargs["work_dir"]
        self.stdout_file = self.kwargs["stdout_file"]
        self.txt_parser = EspressoTXTParser(self.work_dir)
        self.xml_parser = EspressoXMLParser(self.find_xml_file())

    def find_xml_file(self):
        """
        Finds XML file.

        Note: Sternheimer GW XML file of the first process (gw0) is returned if this a Sternheimer GW calculation.

        Returns:
            str
        """
        is_sternheimer_gw = self._is_sternheimer_gw_calculation()
        for root, dirs, files in os.walk(self.work_dir, followlinks=True):
            for file_ in [f for f in files if settings.XML_DATA_FILE == f]:
                file_path = os.path.join(root, file_)
                if not is_sternheimer_gw or (is_sternheimer_gw and settings.STERNHEIMER_GW0_DIR_PATTERN in file_path):
                    return file_path

    def total_energy(self):
        """
        Returns total energy.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.total_energy
        """
        return self.txt_parser.total_energy(self._get_file_content(self.stdout_file))

    def fermi_energy(self):
        """
        Returns fermi energy.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.fermi_energy
        """
        return self.xml_parser.fermi_energy()

    def nspins(self):
        """
        Returns the number of spins.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.nspins
        """
        return self.xml_parser.nspins()

    def _is_sternheimer_gw_calculation(self):
        """
        Checks whether this is a Sternheimer GW calculation.

        The calculation is considered Sternheimer GW if "SternheimerGW" is written on top of the stdout file.

        NOTE: DO NOT READ THE WHOLE FILE INTO MEMORY AS IT IS BIG.

        Returns:
             bool
        """
        if os.path.exists(self.stdout_file):
            with open(self.stdout_file, "r") as f:
                for index, line in enumerate(f):
                    if index > 50: return False
                    if settings.STERNHEIMER_GW_TITLE in line:
                        return True

    def eigenvalues_at_kpoints(self):
        """
        Returns eigenvalues for all kpoints.

        NOTE: eigenvalues are extracted from Sternheimer GW stdout file if this is a Sternheimer GW calculation.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.eigenvalues_at_kpoints
        """
        if self._is_sternheimer_gw_calculation():
            text = self._get_file_content(self.stdout_file)
            inverse_reciprocal_lattice_vectors = self.xml_parser.get_inverse_reciprocal_lattice_vectors()
            return self.txt_parser.eigenvalues_at_kpoints_from_sternheimer_gw_stdout(text,
                                                                                     inverse_reciprocal_lattice_vectors)
        else:
            return self.xml_parser.eigenvalues_at_kpoints()

    def ibz_k_points(self):
        """
        Returns ibz_k_points.

        Note:
            The function assumes that kpoints extracted from parsed source are inside the irreducible wedge of the
            Brillouin zone. Without checking whether it is the case or not.

        Reference:
            func: express.parsers.mixins.reciprocal.ReciprocalDataMixin.ibz_k_points
        """
        return np.array([eigenvalueData["kpoint"] for eigenvalueData in self.eigenvalues_at_kpoints()])

    def dos(self):
        """
        Returns density of states.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.dos
        """
        return self.txt_parser.dos()

    def initial_basis(self):
        """
        Returns initial basis.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.initial_basis
        """
        return self.txt_parser.initial_basis(self._get_file_content(self.stdout_file))

    def initial_lattice_vectors(self):
        """
        Returns initial lattice vectors.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.initial_lattice_vectors
        """
        return self.txt_parser.initial_lattice_vectors(self._get_file_content(self.stdout_file))

    def final_basis(self):
        """
        Returns final basis.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.basis
        """
        return self.xml_parser.final_basis()

    def final_lattice_vectors(self):
        """
        Returns final lattice.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.lattice_vectors
        """
        return self.xml_parser.final_lattice_vectors(reciprocal=False)

    def convergence_electronic(self):
        """
        Extracts convergence electronic.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.convergence_electronic
        """
        return self.txt_parser.convergence_electronic(self._get_file_content(self.stdout_file))

    def convergence_ionic(self):
        """
        Returns convergence ionic.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.convergence_ionic
        """
        return self.txt_parser.convergence_ionic(self._get_file_content(self.stdout_file))

    def stress_tensor(self):
        """
        Returns stress tensor.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.stress_tensor
        """
        return self.txt_parser.stress_tensor(self._get_file_content(self.stdout_file))

    def pressure(self):
        """
        Returns pressure.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.pressure
        """
        return self.txt_parser.pressure(self._get_file_content(self.stdout_file))

    def total_force(self):
        """
        Returns total force.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.total_force
        """
        return self.txt_parser.total_force(self._get_file_content(self.stdout_file))

    def atomic_forces(self):
        """
        Returns forces that is exerted on each atom by its surroundings.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.atomic_forces
        """
        return self.txt_parser.atomic_forces(self._get_file_content(self.stdout_file))

    def total_energy_contributions(self):
        """
        Extracts total energy contributions.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.total_energy_contributions
        """
        return self.txt_parser.total_energy_contributions(self._get_file_content(self.stdout_file))

    def zero_point_energy(self):
        """
        Returns zero point energy.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.zero_point_energy
        """
        return self.txt_parser.zero_point_energy(self._get_file_content(self.stdout_file))

    def phonon_dos(self):
        """
        Returns phonon dos.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.phonon_dos
        """
        return self.txt_parser.phonon_dos()

    def phonon_dispersions(self):
        """
        Returns phonon dispersions.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.phonon_dispersions
        """
        return self.txt_parser.phonon_dispersions()

    def _find_neb_dat_file(self):
        neb_path_file = find_file(NEB_PATH_FILE_SUFFIX, self.work_dir)
        if neb_path_file: return "{}.dat".format(neb_path_file[:neb_path_file.rfind(".")])

    def reaction_coordinates(self):
        """
        Returns reaction coordinates.

        Returns:
             list
        """
        neb_dat_file = self._find_neb_dat_file()
        return self.txt_parser.reaction_coordinates(self._get_file_content(neb_dat_file))

    def reaction_energies(self):
        """
        Returns reaction energies.

        Returns:
             list
        """
        neb_dat_file = self._find_neb_dat_file()
        return self.txt_parser.reaction_energies(self._get_file_content(neb_dat_file))

    def _get_esm_file(self):
        return find_file(".esm1", self.work_dir)

    def potential_profile(self):
        return self.txt_parser.potential_profile(self._get_file_content(self._get_esm_file()))

    def charge_density_profile(self):
        return self.txt_parser.charge_density_profile(self._get_file_content(self._get_esm_file()))

    def _is_pw_scf_output_file(self, path):
        """
        Checks whether the given file is PWSCF output file.

        The file is considered PWSCF output file if "Program PWSCF" is written on top of the file.

        NOTE: DO NOT READ THE WHOLE FILE INTO MEMORY AS IT COULD BE BIG.

        Returns:
             bool
        """
        if os.path.exists(path):
            with open(path, "r") as f:
                for index, line in enumerate(f):
                    if index > 50: return False
                    if settings.PWSCF_OUTPUT_FILE_REGEX in line:
                        return True

    def _find_pw_scf_output_files(self):
        pw_scf_output_files = []
        for root, dirs, files in os.walk(self.work_dir, followlinks=True):
            for file in files:
                path = os.path.join(root, file)
                if self._is_pw_scf_output_file(path):
                    pw_scf_output_files.append(path)
        return pw_scf_output_files

    def initial_structure_strings(self):
        structures = []
        for pw_scf_output_file in self._find_pw_scf_output_files():
            try:
                basis = self.txt_parser.initial_basis(self._get_file_content(pw_scf_output_file))
                lattice = self.txt_parser.initial_lattice_vectors(self._get_file_content(pw_scf_output_file))
                structures.append(lattice_basis_to_poscar(lattice, basis))
            except:
                raise
        return structures

    def final_structure_strings(self):
        structures = []
        for pw_scf_output_file in self._find_pw_scf_output_files():
            try:
                basis = self.txt_parser.final_basis(self._get_file_content(pw_scf_output_file))
                lattice = self.txt_parser.final_lattice_vectors(self._get_file_content(pw_scf_output_file))
                structures.append(lattice_basis_to_poscar(lattice, basis))
            except:
                pass
        return structures



from express.properties.non_scalar.two_dimensional_plot import TwoDimensionalPlotProperty


class ChargeDensityProfile(TwoDimensionalPlotProperty):
    """
    Charge Density profile.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(ChargeDensityProfile, self).__init__(name, parser, *args, **kwargs)
        charge_density_profile = self.parser.charge_density_profile()
        self.xDataArray = charge_density_profile[0]
        self.yDataSeries = [charge_density_profile[1]]



from tests.unit import UnitTestBase
from tests.fixtures.data import TOTAL_ENERGY_CONTRIBUTIONS_RAW_DATA
from express.properties.non_scalar.total_energy_contributions import TotalEnergyContributions

TOTAL_ENERGY_CONTRIBUTIONS = {
    "name": "total_energy_contributions",
    "ewald": {
        "name": "ewald",
        "value": 128376.45871064
    },
    "hartree": {
        "name": "hartree",
        "value": -145344.66902862
    },
    "exchangeCorrelation": {
        "name": "exchange_correlation",
        "value": 41.63693035
    },
    "units": "eV"
}


class TotalEnergyContributionsTest(UnitTestBase):
    def setUp(self):
        super(TotalEnergyContributionsTest, self).setUp()

    def tearDown(self):
        super(TotalEnergyContributionsTest, self).setUp()

    def test_total_energy_contributions(self):
        parser = self.get_mocked_parser("total_energy_contributions", TOTAL_ENERGY_CONTRIBUTIONS_RAW_DATA)
        property_ = TotalEnergyContributions("total_energy_contributions", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), TOTAL_ENERGY_CONTRIBUTIONS)



import os
import yaml
import unittest
import numpy as np


class TestBase(unittest.TestCase):
    """
    Base class for express tests.
    """

    def setUp(self):
        super(TestBase, self).setUp()
        self.rootDir = os.path.dirname(__file__)

    def tearDown(self):
        super(TestBase, self).tearDown()

    def getManifest(self):
        """
        Returns test's manifest.

        Returns:
            dict
        """
        with open(os.path.join(self.rootDir, "manifest.yaml")) as f:
            return yaml.load(f.read(), Loader=yaml.FullLoader)[self._testMethodName]

    def assertDeepAlmostEqual(self, expected, actual, *args, **kwargs):
        """
        Asserts that two complex structures have almost equal contents. Compares lists, dicts and tuples recursively.
        Checks numeric values using assertAlmostEqual() and checks all other values with assertEqual(). Accepts
        additional positional and keyword arguments and passes those intact to assertAlmostEqual().

        Notes:
            Based on: http://stackoverflow.com/a/23550280

        Args:
            expected (dict|list|tuple): expected complex object.
            actual (dict|list|tuple): actual complex object.
        """
        is_root = '__trace' not in kwargs
        trace = kwargs.pop('__trace', 'ROOT')
        try:
            if isinstance(expected, (int, float, complex)):
                self.assertAlmostEqual(expected, actual, *args, **kwargs)
            elif isinstance(expected, (str)):
                self.assertEqual(expected, actual)
            elif isinstance(expected, (list, tuple, np.ndarray)):
                self.assertEqual(len(expected), len(actual))
                for index in range(len(expected)):
                    v1, v2 = expected[index], actual[index]
                    self.assertDeepAlmostEqual(v1, v2, __trace=repr(index), *args, **kwargs)
            elif isinstance(expected, dict):
                self.assertEqual(set(expected), set(actual))
                for key in expected:
                    self.assertDeepAlmostEqual(expected[key], actual[key], __trace=repr(key), *args, **kwargs)
        except AssertionError as exc:
            exc.__dict__.setdefault('traces', []).append(trace)
            if is_root:
                trace = ' -> '.join(reversed(exc.traces))
                exc = AssertionError("%s\nTRACE: %s" % (str(exc), trace))
            raise exc



import os
import numpy as np

from express.parsers import BaseParser
from express.parsers.apps.vasp import settings
from express.parsers.mixins.ionic import IonicDataMixin
from express.parsers.apps.vasp.formats.txt import VaspTXTParser
from express.parsers.apps.vasp.formats.xml import VaspXMLParser
from express.parsers.mixins.reciprocal import ReciprocalDataMixin
from express.parsers.mixins.electronic import ElectronicDataMixin
from express.parsers.utils import find_file, find_files_by_name_substring
from express.parsers.apps.vasp.settings import NEB_DIR_PREFIX, NEB_STD_OUT_FILE


class VaspParser(BaseParser, IonicDataMixin, ElectronicDataMixin, ReciprocalDataMixin):
    """
    Vasp parser class.
    """

    def __init__(self, *args, **kwargs):
        super(VaspParser, self).__init__(*args, **kwargs)
        self.work_dir = self.kwargs["work_dir"]
        self.stdout_file = self.kwargs["stdout_file"]
        self.txt_parser = VaspTXTParser(self.work_dir)
        self.xml_parser = VaspXMLParser(find_file(settings.XML_DATA_FILE, self.work_dir))

    def _get_outcar_content(self):
        """
        Returns the content of OUTCAR file.

        Returns:
            str
        """
        outcar_content = ''
        outcar_path = os.path.join(self.work_dir, "OUTCAR")
        if os.path.exists(outcar_path):
            with open(outcar_path) as f:
                outcar_content = f.read()
        return outcar_content

    def total_energy(self):
        """
        Returns total energy.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.total_energy
        """
        return self.txt_parser.total_energy(self._get_file_content(self.stdout_file))

    def fermi_energy(self):
        """
        Returns fermi energy.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.fermi_energy
        """
        return self.xml_parser.fermi_energy()

    def nspins(self):
        """
        Returns the number of spins.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.nspins
        """
        return self.xml_parser.nspins()

    def eigenvalues_at_kpoints(self):
        """
        Returns eigenvalues for all kpoints.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.eigenvalues_at_kpoints
        """
        return self.xml_parser.eigenvalues_at_kpoints()

    def ibz_k_points(self):
        """
        Returns ibz_k_points.

        Note:
            The function assumes that kpoints extracted from parsed source are inside the irreducible wedge of the
            Brillouin zone. Without checking whether it is the case or not.

        Reference:
            func: express.parsers.mixins.reciprocal.ReciprocalDataMixin.ibz_k_points
        """
        return np.array([eigenvalueData["kpoint"] for eigenvalueData in self.eigenvalues_at_kpoints()])

    def dos(self):
        """
        Returns density of states.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.dos
        """
        return self.xml_parser.dos(combined=True)

    def final_basis(self):
        """
        Returns final basis.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.final_basis
        """
        return self.xml_parser.final_basis()

    def final_lattice_vectors(self):
        """
        Returns final lattice vectors.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.final_lattice_vectors
        """
        return self.xml_parser.final_lattice_vectors()

    def convergence_electronic(self):
        """
        Extracts convergence electronic.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.convergence_electronic
        """
        outcar = self._get_outcar_content()
        stdout = self._get_file_content(self.stdout_file)
        try:
            atom_names = self.xml_parser.atom_names()
        except:
            print("atom_names can not be extracted")
            atom_names = []
        return self.txt_parser.convergence_electronic(outcar, stdout, atom_names)

    def convergence_ionic(self):
        """
        Extracts convergence ionic.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.convergence_ionic
        """
        outcar = self._get_outcar_content()
        stdout = self._get_file_content(self.stdout_file)
        return self.txt_parser.convergence_ionic(outcar, stdout, self.xml_parser.atom_names())

    def stress_tensor(self):
        """
        Returns stress tensor.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.stress_tensor
        """
        return self.xml_parser.stress_tensor()

    def pressure(self):
        """
        Returns pressure.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.pressure
        """
        return self.txt_parser.pressure(self._get_outcar_content())

    def total_force(self):
        """
        Returns total force.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.total_force
        """
        return self.txt_parser.total_force(self._get_outcar_content())

    def atomic_forces(self):
        """
        Returns forces that is exerted on each atom by its surroundings.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.atomic_forces
        """
        return self.xml_parser.atomic_forces()

    def total_energy_contributions(self):
        """
        Extracts total energy contributions.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.total_energy_contributions
        """
        return self.txt_parser.total_energy_contributions(self._get_outcar_content())

    def zero_point_energy(self):
        """
        Returns zero point energy.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.zero_point_energy
        """
        return self.txt_parser.zero_point_energy(self._get_outcar_content())

    def magnetic_moments(self):
        """
        Returns magnetic moments.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.magnetic_moments
        """
        return self.txt_parser.magnetic_moments(os.path.join(self.work_dir, "OUTCAR"))

    def _get_neb_image_dirs(self, prefix="0"):
        """
        Returns the sorted list of NEB image directories.

        Args:
            prefix (str): image directory prefix.

        Returns:
             list[str]
        """
        paths = []
        for root, dirs, files in os.walk(self.work_dir):
            for dir_ in [d for d in dirs if str(d).startswith(prefix)]:
                path = os.path.join(root, dir_)
                paths.append({"path": path, "index": int(dir_)})
            break
        return [p["path"] for p in sorted(paths, key=lambda p: p["index"])]

    def _get_neb_image_stdout_files(self, prefix="0", output_file="stdout"):
        """
        Returns the paths to NEB image standard output files.

        Note: only image 01 writes to the usual stdout file, other images write to `stdout` inside image directory.

        Args:
            prefix (str): image directory prefix.
            output_file (str): output file name.

        Returns:
             list[str]
        """
        files = []
        for path in self._get_neb_image_dirs(prefix):
            file_ = self.stdout_file if os.path.basename(path) == "01" else os.path.join(path, output_file)
            files.append(file_)
        return files

    def reaction_energies(self):
        """
        Returns reaction energies.

        Returns:
             list
        """
        energies = []
        for path in self._get_neb_image_stdout_files(NEB_DIR_PREFIX, NEB_STD_OUT_FILE):
            energies.append(self.txt_parser.total_energy(self._get_file_content(path)))
        return [energies[i] - energies[0] for i in range(len(energies))]

    def reaction_coordinates(self):
        """
        Returns reaction coordinates.

        Note: it is assumed that initial and final directories contain CONTCAR file.

        Returns:
             list
        """
        structures = []
        for path in self._get_neb_image_dirs():
            with open(os.path.join(path, "CONTCAR")) as f:
                structures.append(f.read())
        return self.reaction_coordinates_from_poscars(structures)

    def initial_structure_strings(self):
        structures = []
        for poscar in find_files_by_name_substring("POSCAR", self.work_dir):
            with open(poscar) as f:
                structures.append(f.read())
        return structures

    def final_structure_strings(self):
        structures = []
        for poscar in find_files_by_name_substring("CONTCAR", self.work_dir):
            with open(poscar) as f:
                structures.append(f.read())
        return structures



from math import pi
from munch import Munch


class Constant(object):
    """
    Constants from Konrad Hinsen's PhysicalQuantities module (1986 CODATA).
    """
    _c = 299792458.  # speed of light, m/s
    _mu0 = 4.e-7 * pi  # permeability of vacuum
    _eps0 = 1 / _mu0 / _c ** 2  # permittivity of vacuum
    _Grav = 6.67259e-11  # gravitational constant
    _hplanck = 6.6260755e-34  # Planck constant, J s
    _hbar = _hplanck / (2 * pi)  # Planck constant / 2pi, J s
    _e = 1.60217733e-19  # elementary charge
    _me = 9.1093897e-31  # electron mass
    _mp = 1.6726231e-27  # proton mass
    _Nav = 6.0221367e23  # Avogadro number
    _k = 1.380658e-23  # Boltzmann constant, J/K
    _amu = 1.6605402e-27  # atomic mass unit, kg
    BOHR = 4e10 * pi * _eps0 * _hbar ** 2 / _me / _e ** 2  # Bohr radius
    eV = 1.0
    HARTREE = _me * _e ** 3 / 16 / pi ** 2 / _eps0 ** 2 / _hbar ** 2
    RYDBERG = 0.5 * HARTREE
    Ry = RYDBERG
    Ha = HARTREE
    kJ = 1000.0 / _e
    kcal = 4.184 * kJ
    cm_inv_to_ev = 0.00012398  # cm^-1 to eV
    ry_bohr_to_eV_A = 25.71104309541616  # or RYDBERG / BOHR

GENERAL_REGEX = Munch()
GENERAL_REGEX.update({
    'double_number': r'[-+]?\d*\.\d+(?:[eE][-+]?\d+)?',
    'int_number': r'[+-]?\d+'
})

ASE_FORMATS = {
    "poscar": "vasp",
    "cif": "cif",
    "espresso-in": "espresso-in"
}



from express.properties.non_scalar import NonScalarProperty


class TotalEnergyContributions(NonScalarProperty):
    """
    Total energy contribution factors.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(TotalEnergyContributions, self).__init__(name, parser, *args, **kwargs)

    def _serialize(self):
        data = {
            'name': self.name,
            'units': self.manifest["defaults"]["units"]
        }
        data.update(self.parser.total_energy_contributions())
        return data



import json
import os

current_path = os.path.dirname(__file__)
path = os.path.join(current_path, "test-001")
predict_workflow = "workflow_predict_after_conversion.JSON"
train_workflow = "workflow_train.JSON"

# Get both workflows
with open(os.path.join(path, train_workflow), "r") as file_pointer:
    WORKFLOW_TRAIN = json.load(file_pointer)

with open(os.path.join(path, predict_workflow), "r") as file_pointer:
    WORKFLOW_PREDICT = json.load(file_pointer)

NAME = "workflow:pyml_predict"
PARSER = None
ARGS = []

# Get basenames to mock for the os.listdirs call in test
_subworkflow_units = WORKFLOW_PREDICT["subworkflows"][0]["units"]
_download_from_s3_unit = next(filter(lambda i: i["name"] == "Fetch Trained Model as file", _subworkflow_units))
_download_from_s3_inputs = _download_from_s3_unit["input"]
MOCK_BASENAMES = [item["basename"] for item in _download_from_s3_inputs]

# Determine working directory information
_sample_object_data = _download_from_s3_inputs[0]["objectData"]
_job_context_abspath = os.path.dirname(_sample_object_data["NAME"])
CONTEXT_DIR_RELATIVE_PATH = os.path.basename(_job_context_abspath)

WORK_DIR = os.path.dirname(_job_context_abspath)
UPLOAD_DIR = WORK_DIR

# Object Storage Info
_object_storage_container = _sample_object_data["CONTAINER"]
_object_storage_provider = _sample_object_data["PROVIDER"]
_object_storage_region = _sample_object_data["REGION"]
OBJECT_STORAGE_DATA = {
    "CONTAINER": _object_storage_container,
    "NAME": "",
    "PROVIDER": _object_storage_provider,
    "REGION": _object_storage_region
}



import re
import numpy as np
from pymatgen.io.vasp import Outcar

from express.parsers.apps.vasp import settings
from express.parsers.formats.txt import BaseTXTParser


class VaspTXTParser(BaseTXTParser):
    """
    Vasp text parser class.
    """

    def __init__(self, work_dir):
        super(VaspTXTParser, self).__init__(work_dir)

    def ibz_kpoints(self, text, space):
        """
        Extracts kpoints coordinates in cartesian space.

        Args:
            text (str): text to extract data from.
            space (str): kpoints coordinate space, either crystal or cartesian.

        Returns:
            ndarray
        """
        text_range = {
            'cartesian': {
                'start': 'Following cartesian coordinates',
                'end': 'Dimension of arrays'
            },
            'crystal': {
                'start': 'Following reciprocal coordinates',
                'end': 'Following cartesian coordinates'
            }
        }
        start_index = text.find(text_range[space]['start'])
        end_index = text.find(text_range[space]['end'])
        ibz_kpts = re.findall(settings.REGEX["ibz_kpoints"]["regex"], text[start_index:end_index])
        ibz_kpts = [[float(x) for x in kp] for kp in ibz_kpts]
        return np.array(ibz_kpts)

    def total_energy(self, text):
        """
        Extracts total energy.

        Args:
            text (str): text to extract data from.

        Returns:
             float
        """
        return self._general_output_parser(text, **settings.REGEX["total_energy"])

    def _lattice_convergence(self, text):
        """
        Extracts lattice convergence values computed in each BFGS step.

        Args:
            text (str): text to extract data from.

        Returns:
            list

        Example:
            [
                {
                    'vectors': {
                        'a': [-0.561154473, -0.000000000, 0.561154473],
                        'b': [-0.000000000, 0.561154473, 0.561154473],
                        'c': [-0.561154473, 0.561154473, 0.000000000],
                        'alat': 9.44858082
                    }
                 },
                ...
                {
                    'vectors': {
                        'a': [-0.561154473, -0.000000000, 0.561154473],
                        'b': [-0.000000000, 0.561154473, 0.561154473],
                        'c': [-0.561154473, 0.561154473, 0.000000000],
                        'alat': 9.44858082
                    }
                 }
             ]
        """

        lattices = []
        match = re.findall(settings.REGEX["lattice_vectors"]["regex"], text)
        if match:
            for lattice in match:
                lattice = [float(_) for _ in lattice]
                lattices.append({
                    'vectors': {
                        'a': lattice[0:3],
                        'b': lattice[3:6],
                        'c': lattice[6:9],
                        'alat': 1.0  # abc vectors are expected in absolute units (eg. bohr)
                    }
                })
        return lattices

    def _basis_convergence(self, text, atom_names):
        """
        Extracts convergence bases computed in each BFGS step.

        Args:
            text (str): text to extract data from.
            atom_names (list): list of atom names.

        Returns:
            list

        Example:
            [
                {
                    'units': 'crystal',
                    'elements': [{'id': 1, 'value': 'Si'}, {'id': 2, 'value': 'Si'}],
                    'coordinates': [{'id': 1, 'value': [0.0, 0.0, 0.0]}, {'id': 2, 'value': [0.0, 0.0, 0.0]}]
                 },
                ...
                {
                    'units': 'crystal',
                    'elements': [{'id': 1, 'value': 'Si'}, {'id': 2, 'value': 'Si'}],
                    'coordinates': [{'id': 1, 'value': [0.0, 0.0, 0.0]}, {'id': 2, 'value': [0.0, 0.0, 0.0]}]
                 }
            ]
        """
        results = []
        matches = re.findall(settings.REGEX["ion_positions_block"]["regex"], text, re.DOTALL | re.MULTILINE)
        if matches:
            for match in matches:
                ions = re.findall(settings.REGEX["basis_vectors"]["regex"], match)
                results.append({
                    "units": "angstrom",
                    "elements": [{"id": idx, "value": atom_names[idx]} for idx in range(len(ions))],
                    "coordinates": [{"id": idx, "value": [float(x) for x in ion]} for idx, ion in enumerate(ions)]
                })
            return results

    def convergence_electronic(self, outcar, stdout, atom_names):
        """
        Extracts convergence electronic.
            1. Extract all energies (from dE column) along with the corresponding step [(1, 0.69948E+04), (2, -0.73973E+04)]
            2. Group the energies for each ionic step

        Sample input:
                   N       E                     dE             d eps       ncg     rms          rms(c)
            DAV:   1     0.699475439520E+04    0.69948E+04   -0.37054E+05   920   0.140E+03
            DAV:   2    -0.402547717878E+03   -0.73973E+04   -0.71440E+04  1084   0.442E+02
            DAV:   3    -0.105567783952E+04   -0.65313E+03   -0.64960E+03  1160   0.125E+02
            DAV:   4    -0.110887818556E+04   -0.53200E+02   -0.53162E+02  1560   0.284E+01
            DAV:   5    -0.111382548259E+04   -0.49473E+01   -0.49469E+01  1240   0.562E+00    0.137E+02
                1 T=  2001. E= -.89713765E+03 F= -.92170983E+03 E0= -.92172095E+03  EK= 0.24572E+02

                   N       E                     dE             d eps       ncg     rms          rms(c)
            DAV:   1    -0.916206954626E+03    0.55030E+01   -0.19221E+03   920   0.750E+01    0.931E+00
            DAV:   2    -0.920149095159E+03   -0.39421E+01   -0.47144E+01  1072   0.109E+01    0.625E+00
            DAV:   3    -0.919835295300E+03    0.31380E+00   -0.27646E+00  1288   0.256E+00    0.367E+00
            DAV:   4    -0.919705445969E+03    0.12985E+00   -0.93950E-01  1128   0.186E+00    0.145E+00
                2 T=  1843. E= -.89706481E+03 F= -.91969309E+03 E0= -.91970717E+03  EK= 0.22628E+02

        Args:
            outcar (str): OUTCAR content.
            stdout (str): stdout content.
            atom_names (list): list of atoms.

        Returns:
             list[list]
        """
        data = []
        matches = self._general_output_parser(stdout, **settings.REGEX["convergence_electronic"])
        first_step_indices = [ind for ind, elm in enumerate(matches) if int(elm[0]) == 1]
        for ind, first_step_index in enumerate(first_step_indices):
            if ind + 1 == len(first_step_indices):
                energies = matches[first_step_index:]
            else:
                energies = matches[first_step_index:first_step_indices[ind + 1]]
            data.append([energy[1] for energy in energies])  # strip out the step numbers
        return data

    def convergence_ionic(self, outcar, stdout, atom_names):
        """
        Extracts convergence ionic.

        Args:
            outcar (str): OUTCAR content.
            stdout (str): stdout content.
            atom_names (list): list of atoms.

        Returns:
             list[dict]
        """
        data = []
        convergence_electronic_data = self.convergence_electronic(outcar, stdout, atom_names)
        for ind, energies in enumerate(convergence_electronic_data):
            data.append({
                "energy": sum([sum(e) for e in convergence_electronic_data[0:ind + 1]]),
                "electronic": {
                    "units": "eV",
                    "data": energies
                },
            })

        if not data: return []

        lattice_convergence = self._lattice_convergence(outcar)
        basis_convergence = self._basis_convergence(outcar, atom_names)
        data = data[0:len(lattice_convergence)]  # strip out the last non-complete step
        if not data: return []
        for idx, structure in enumerate(zip(lattice_convergence, basis_convergence)):
            data[idx].update({
                'structure': {
                    'lattice': structure[0],
                    'basis': structure[1]
                }
            })

        return data

    def pressure(self, text):
        """
        Extracts pressure.

        Args:
            text (str): text to extract data from.

        Returns:
            float
        """
        return self._general_output_parser(text, **settings.REGEX["pressure"])

    def total_force(self, text):
        """
        Extracts total force.

        Args:
            text (str): text to extract data from.

        Returns:
            float
        """
        total_force = self._general_output_parser(text, **settings.REGEX['total_force'])
        return np.sqrt(np.sum(np.square(total_force)))

    def total_energy_contributions(self, text):
        """
        Extracts total energy contributions.

        Args:
            text (str): text to extract data from.

        Returns:
            dict
        """
        energy_contributions = {}
        for contribution in settings.TOTAL_ENERGY_CONTRIBUTIONS:
            value = self._general_output_parser(text, **settings.TOTAL_ENERGY_CONTRIBUTIONS[contribution])
            if value:
                energy_contributions.update({contribution: {
                    'name': contribution,
                    'value': np.sqrt(np.sum(np.square(value)))
                }})
        return energy_contributions

    def zero_point_energy(self, text):
        """
        Extracts zero point energy.

        Returns:
             float
        """
        data = self._general_output_parser(text, **settings.REGEX['zero_point_energy'])
        if len(data): return sum(data) / 2 / 1000

    def magnetic_moments(self, outcar):
        """
        Extracts magnetic moments.

        Returns:
             list
        """
        mag = Outcar(outcar).magnetization
        return [[0, 0, ion['tot']] if isinstance(ion['tot'], float) else ion['tot'].moment.tolist() for ion in mag]



from tests.unit import UnitTestBase
from tests.fixtures.data import PHONON_DOS_RAW_DATA
from express.properties.non_scalar.two_dimensional_plot.phonon_dos import PhononDOS

PHONON_DOS = {
    "yDataSeries": [
        [
            0.0,
            1.7269000451847205e-08,
            6.90749999421314e-08
        ]
    ],
    "xDataArray": [
        [
            -313.8999938964844,
            -312.8999938964844,
            -311.8999938964844
        ]
    ],
    "name": "phonon_dos",
    "xAxis": {
        "units": "cm-1",
        "label": "frequency"
    },
    "yAxis": {
        "units": "states/cm-1",
        "label": "Phonon DOS"
    }
}


class PhononDOSTest(UnitTestBase):
    def setUp(self):
        super(PhononDOSTest, self).setUp()

    def tearDown(self):
        super(PhononDOSTest, self).setUp()

    def test_phonon_dos(self):
        parser = self.get_mocked_parser("phonon_dos", PHONON_DOS_RAW_DATA)
        property_ = PhononDOS("phonon_dos", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), PHONON_DOS)



from tests.unit import UnitTestBase
from express.properties.scalar.total_force import TotalForce

TOTAL_FORCE = {
    "units": "eV/angstrom",
    "name": "total_force",
    "value": 1
}


class TotalForceTest(UnitTestBase):
    def setUp(self):
        super(TotalForceTest, self).setUp()

    def tearDown(self):
        super(TotalForceTest, self).setUp()

    def test_total_force(self):
        parser = self.get_mocked_parser("total_force", 1)
        property_ = TotalForce("total_force", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), TOTAL_FORCE)



from express.parsers.settings import GENERAL_REGEX

COMMON_REGEX = r"{}\s+[=:<>]\s*([-+]?\d*\.?\d*([Ee][+-]?\d+)?)"
DOUBLE_REGEX = GENERAL_REGEX.double_number
NWCHEM_OUTPUT_FILE_REGEX = "Northwest Computational Chemistry Package"

REGEX = {
    "total_energy": {
        "regex": COMMON_REGEX.format("Total DFT energy"),
        "occurrences": -1,
        "output_type": "float"
    }
}
TOTAL_ENERGY_CONTRIBUTIONS = {
    "one_electron": {
        "regex": COMMON_REGEX.format("One electron energy"),
        "occurrences": -1,
        "output_type": "float"
    },
    "coulomb": {
        "regex": COMMON_REGEX.format("Coulomb Energy"),
        "occurrences": -1,
        "output_type": "float"
    },
    "exchange_correlation": {
        "regex": COMMON_REGEX.format("Exchange-Corr. energy"),
        "occurrences": -1,
        "output_type": "float"
    },
    "nuclear_repulsion": {
        "regex": COMMON_REGEX.format("Nuclear repulsion energy"),
        "occurrences": -1,
        "output_type": "float"
    }
}



from express.parsers.settings import GENERAL_REGEX

PDOS_TOT_FILE = "pdos_tot"
NEB_PATH_FILE_SUFFIX = ".path"
XML_DATA_FILE = "data-file.xml"
PHONON_DOS_FILE = "phonon_dos.out"
PHONON_MODES_FILE = "normal_modes.out"

COMMON_REGEX = r"{0}\s+[=:<>]\s*([-+]?\d*\.?\d*([Ee][+-]?\d+)?)"
DOUBLE_REGEX = GENERAL_REGEX.double_number

STERNHEIMER_GW0_DIR_PATTERN = "/_gw0/"
STERNHEIMER_GW_TITLE = "SternheimerGW"
PWSCF_OUTPUT_FILE_REGEX = "Program PWSCF"

REGEX = {
    "total_energy": {
        "regex": COMMON_REGEX.format("total energy"),
        "start_flag": "!",
        "occurrences": -1,
        "output_type": "float"
    },
    "pdos_file": {
        "regex": r'.*\.pdos_atm#(?P<atom_num>\d+)\((?P<atom_name>\w+)\)'
                 r'_wfc#(?P<orbit_num>\d+)\((?P<orbit_symbol>\w)\)',
    },
    "convergence_electronic": {
        "regex": r"estimated scf accuracy\s+<\s+({0})".format(DOUBLE_REGEX),
        "output_type": "float",
    },
    "convergence_ionic_blocks": {
        "regex": r"\s+Self-consistent Calculation.+?\n(.+?)\s+convergence has been achieved"
    },
    "convergence_ionic_energies": {
        "regex": r"total energy\s+=\s+({0})".format(DOUBLE_REGEX),
        "output_type": "float",
    },
    "bfgs_block": {
        "regex": r"new unit-cell volume.+?Writing output data file",
    },
    "lattice": {
        "regex": (
            r"CELL_PARAMETERS\s+\(angstrom\)"
            r"\s+({0})\s+({0})\s+({0})"
            r"\s+({0})\s+({0})\s+({0})"
            r"\s+({0})\s+({0})\s+({0})"
        ).format(DOUBLE_REGEX)
    },
    "lattice_alat": {
        "regex": (
            r"crystal axes:.*\n"
            r".*(\s+{0})\s+({0})\s+({0}\s+).*\n"
            r".*(\s+{0})\s+({0})\s+({0}\s+).*\n"
            r".*(\s+{0})\s+({0})\s+({0}\s+).*\n"
        ).format(DOUBLE_REGEX)
    },
    "lattice_parameter_alat": {
        "regex": r"lattice parameter \(alat\)\s+=\s+({0})\s+".format(DOUBLE_REGEX),
        "output_type": "float",
    },
    "number_of_atoms": {
        "regex": r"number of atoms/cell\s+=\s+(\d+)",
        "output_type": "int",
    },
    "basis_alat": lambda number_of_atoms: {
        "regex": r".+?\d\s+([A-Z][a-z]?).+?({0})\s+({0})\s+({0}).+?\n".format(DOUBLE_REGEX),
        "occurrences": number_of_atoms,
        "output_type": "str",
        "match_groups": [1, 2, 3, 4]
    },
    "ion_position": {
        "regex": r"([A-Z][a-z]?)\s+({0})\s+({0})\s+({0})".format(DOUBLE_REGEX)
    },
    "stress_tensor": {
        "regex": r"^\s*{0}\s+{0}\s+{0} +({0})\s+({0})\s+({0})".format(DOUBLE_REGEX),
        "start_flag": "Forces acting on atoms",
        "occurrences": 3,
        "output_type": "float",
        "match_groups": [1, 2, 3]
    },
    "pressure": {
        "regex": r"\s*total\s+stress\s+\(Ry/bohr\*\*3\)\s*\(kbar\)\s*P=\s*({0})".format(DOUBLE_REGEX),
        "start_flag": "Forces acting on atoms",
        "occurrences": -1,
        "output_type": "float"
    },
    "total_force": {
        "regex": COMMON_REGEX.format("Total force"),
        "start_flag": "Total force",
        "occurrences": -1,
        "output_type": "float"
    },
    "forces_on_atoms": {
        "regex": r"^\s*atom\s+\d+\s+type\s+\d+\s+force\s+=\s+({0})\s+({0})\s+({0})".format(DOUBLE_REGEX),
        "start_flag": "Forces acting on atoms",
        "occurrences": 0,
        "output_type": "float",
        "match_groups": [1, 2, 3]
    },
    "zero_point_energy": {
        "regex": r"freq\s\(\s+\d+\)\s+\=\s+\d+\.\d+\s+\[THz\]\s+\=\s+({0})\s+\[cm\-1\]".format(DOUBLE_REGEX),
        "start_flag": "Diagonalizing the dynamical matrix",
        "output_type": "float"
    },
    'phonon_frequencies': {
        "regex": r'freq\s\(\s+\d+\)\s+\=\s+-*\d+\.\d+\s+\[THz\]\s+\=\s+({})\s+\[cm\-1\]'.format(DOUBLE_REGEX)
    },
    'qpoints': {
        "regex": r'q\s+\=\s+({0})\s+({0})\s+({0})'.format(DOUBLE_REGEX)
    },
    "reaction_coordinates": {
        "regex": r"^\s+({0})\s+{0}\s+{0}".format(DOUBLE_REGEX),
        "output_type": "float",
    },
    "reaction_energies": {
        "regex": r"^\s+{0}\s+({0})\s+{0}".format(DOUBLE_REGEX),
        "output_type": "float",
    },
    "potential_profile": {
        "regex": r"^\s+({0})\s+{0}\s+({0})\s+({0})\s+({0})".format(DOUBLE_REGEX),
        "occurrences": 0,
        "output_type": "float",
        "match_groups": [1, 2, 3, 4]
    },
    "charge_density_profile": {
        "regex": r"^\s+({0})\s+({0})\s+{0}\s+{0}\s+{0}".format(DOUBLE_REGEX),
        "occurrences": 0,
        "output_type": "float",
        "match_groups": [1, 2]
    },
    "sternheimer_gw_kpoint": {
        "regex": r"^\s+GWKpoint cart :\s+({0})\s+({0})\s+({0})".format(DOUBLE_REGEX),
        "occurrences": 0,
        "output_type": "float",
        "match_groups": [1, 2, 3]
    },
    "sternheimer_gw_eigenvalues": {
        "regex": r"^\s+GW qp energy \(eV\)(.*)",
        "occurrences": 0,
        "output_type": "str",
    },
}

TOTAL_ENERGY_CONTRIBUTIONS = {
    "harris_foulkes": {
        "regex": COMMON_REGEX.format("Harris-Foulkes estimate"),
        "start_flag": "!",
        "occurrences": -1,
        "output_type": "float"
    },
    "one_electron": {
        "regex": COMMON_REGEX.format("one-electron contribution"),
        "start_flag": "!",
        "occurrences": -1,
        "output_type": "float"
    },
    "hartree": {
        "regex": COMMON_REGEX.format("hartree contribution"),
        "start_flag": "!",
        "occurrences": -1,
        "output_type": "float"
    },
    "exchange_correlation": {
        "regex": COMMON_REGEX.format("xc contribution"),
        "start_flag": "!",
        "occurrences": -1,
        "output_type": "float"
    },
    "ewald": {
        "regex": COMMON_REGEX.format("ewald contribution"),
        "start_flag": "!",
        "occurrences": -1,
        "output_type": "float"
    },
    "smearing": {
        "regex": COMMON_REGEX.format("smearing contrib\.\s+\(-TS\)"),
        "start_flag": "!",
        "occurrences": -1,
        "output_type": "float"
    }
}



from express.properties.non_scalar.two_dimensional_plot import TwoDimensionalPlotProperty


class PhononDOS(TwoDimensionalPlotProperty):
    """
    Describes the number of vibrational modes per interval of energy.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(PhononDOS, self).__init__(name, parser, *args, **kwargs)
        phonon_dos = self.parser.phonon_dos()
        self.xDataArray = [phonon_dos['frequency']]
        self.yDataSeries = [phonon_dos['total']]



import math
from express.properties.scalar import ScalarProperty


class PNorm(ScalarProperty):
    """
    p-norm property class.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(PNorm, self).__init__(name, parser, *args, **kwargs)
        self.degree = kwargs["degree"]
        ratios = self.parser.elemental_ratios().values()
        self.value = math.pow(sum((math.pow(v, self.degree) for v in ratios)), 1.0 / self.degree) if self.degree else len(ratios)

    def _serialize(self):
        return {
            'name': self.name,
            'value': self.value,
            "degree": self.degree
        }



from mock import MagicMock

from tests.unit import UnitTestBase
from express.properties.non_scalar.two_dimensional_plot.band_structure import BandStructure
from tests.fixtures.data import BAND_STRUCTURE, HSE_EIGENVALUES_AT_KPOINTS, HSE_BAND_STRUCTURE, EIGENVALUES_AT_KPOINTS


class BandStructureTest(UnitTestBase):
    def setUp(self):
        super(BandStructureTest, self).setUp()

    def tearDown(self):
        super(BandStructureTest, self).setUp()

    def test_band_structure(self):
        parser = MagicMock()
        parser.attach_mock(MagicMock(return_value=1), "nspins")
        parser.attach_mock(MagicMock(return_value=EIGENVALUES_AT_KPOINTS), "eigenvalues_at_kpoints")
        property_ = BandStructure("band_structure", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), BAND_STRUCTURE)

    def test_hse_band_structure(self):
        parser = MagicMock()
        parser.attach_mock(MagicMock(return_value=1), "nspins")
        parser.attach_mock(MagicMock(return_value=HSE_EIGENVALUES_AT_KPOINTS), "eigenvalues_at_kpoints")
        property_ = BandStructure("band_structure", parser, remove_non_zero_weight_kpoints=True)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), HSE_BAND_STRUCTURE)



from tests.unit import UnitTestBase
from express.properties.scalar.pressure import Pressure

PRESSURE = {
    "units": "kbar",
    "name": "pressure",
    "value": 1
}


class PressureTest(UnitTestBase):
    def setUp(self):
        super(PressureTest, self).setUp()

    def tearDown(self):
        super(PressureTest, self).setUp()

    def test_pressure(self):
        parser = self.get_mocked_parser("pressure", 1)
        property_ = Pressure("pressure", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), PRESSURE)



from express.properties.scalar import ScalarProperty


class Density(ScalarProperty):
    """
    Density property class.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(Density, self).__init__(name, parser, *args, **kwargs)
        self.value = self.parser.density()



from express.properties.scalar import ScalarProperty


class TotalForce(ScalarProperty):
    """
    Total force in the material.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(TotalForce, self).__init__(name, parser, *args, **kwargs)
        self.value = self.parser.total_force()



import os

from tests.fixtures.structural.references import INCHI_DATA
from tests.integration import IntegrationTestBase
from express.parsers.molecule import MoleculeParser


class TestMoleculeParser(IntegrationTestBase):

    def setUp(self):
        super(IntegrationTestBase, self).setUp()

    def tearDown(self):
        super(IntegrationTestBase, self).setUp()

    @property
    def parser(self):
        manifest = self.getManifest()
        with open(os.path.join(self.rootDir, manifest["structurePath"])) as f:
            kwargs = {
                "cell_type": manifest.get("cell_type", "original"),
                "structure_string": f.read(),
                "structure_format": "poscar"
            }
            return MoleculeParser(**kwargs)

    def test_molecule_inchi(self):
        inchi_long, self.inchi = self.parser.get_inchi()
        self.inchi_key = self.parser.get_inchi_key()
        self.assertEqual(self.inchi["value"], INCHI_DATA["inchi"])
        self.assertEqual(self.inchi_key["value"], INCHI_DATA["inchi_key"])



import ase.io
import rdkit.Chem
from io import StringIO
from typing import Dict, Tuple

from express.parsers.structure import StructureParser
from express.parsers.utils import convert_to_ase_format


class MoleculeParser(StructureParser):
    """
    Molecule parser class.

    Args:
        structure_string (str): structure string.
        structure_format (str): structure format, poscar, cif or espresso-in.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ase_format = convert_to_ase_format(self.structure_format)
        self.inchi_long, self.inchi = self.get_inchi()

    def get_rdkit_mol(self) -> rdkit.Chem.Mol:
        """
        Function to create an RDKit molecule object from a structure string
        """
        ase_pdb = StringIO()

        material_file_string = StringIO(self.structure_string)
        ase_atoms = ase.io.read(material_file_string, format=self.ase_format)

        ase.io.write(ase_pdb, ase_atoms, format="proteindatabank")
        rdkit_mol_object = rdkit.Chem.rdmolfiles.MolFromPDBBlock(ase_pdb.getvalue())
        return rdkit_mol_object

    def get_inchi(self) -> Tuple[str, Dict[str, str]]:
        """
        Function calculates the International Chemical Identifier (InChI) string for a given structure.
        It returns the full InChI string that is calculated along with a shorter notation that omits
        the `InChI=` prefix and is stored as the 'inchi' value.

        Returns:
            Str, Dict

        Example:
            InChI=1S/H2O/h1H2,
            {
                "name": "inchi",
                "value": "1S/H2O/h1H2"
            }
        """

        rdkit_mol_object = self.get_rdkit_mol()

        if rdkit_mol_object is None:
            inchi_short = None
            inchi_long = None
        else:
            inchi_long = rdkit.Chem.inchi.MolToInchi(rdkit_mol_object)
            inchi_short = inchi_long.split("=")[1]
        inchi = {
            "name": "inchi",
            "value": inchi_short
        }
        return inchi_long, inchi

    def get_inchi_key(self) -> Dict[str, str]:
        """
        Function calculates the non-human readable InChI Hash value.

        Returns:
            Dict

        Example:
            InChI Key for H2O
            Dict: {
                      "name": "inchi_key",
                      "value": "XLYOFNOQVPJJNP-UHFFFAOYSA-N"
                  }
        """
        inchi_key_val: str = rdkit.Chem.inchi.InchiToInchiKey(self.inchi_long)
        inchi_key = {
            "name": "inchi_key",
            "value": inchi_key_val
        }
        return inchi_key



from tests.unit import UnitTestBase
from tests.fixtures.data import ATOMIC_FORCES_RAW_DATA
from express.properties.non_scalar.atomic_forces import AtomicForces

ATOMIC_FORCES = {
    "units": "eV/angstrom",
    "name": "atomic_forces",
    "values": [
        {
            "id": 1,
            "value": [
                -3.9e-07,
                -2.4e-07,
                0.0
            ]
        },
        {
            "id": 2,
            "value": [
                3.9e-07,
                2.4e-07,
                0.0
            ]
        }
    ]
}


class AtomicForcesTest(UnitTestBase):
    def setUp(self):
        super(AtomicForcesTest, self).setUp()

    def tearDown(self):
        super(AtomicForcesTest, self).setUp()

    def test_atomic_forces(self):
        parser = self.get_mocked_parser("atomic_forces", ATOMIC_FORCES_RAW_DATA)
        property_ = AtomicForces("atomic_forces", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), ATOMIC_FORCES)



from express.properties import BaseProperty


class ScalarProperty(BaseProperty):
    """
    Base scalar property class.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(ScalarProperty, self).__init__(name, parser, *args, **kwargs)
        self.value = None
        self.units = None

    def _serialize(self):
        return {
            'name': self.name,
            'units': self.manifest["defaults"]["units"],
            'value': self.value,
        }



from express.properties.non_scalar import NonScalarProperty


class ConvergenceElectronic(NonScalarProperty):
    """
    Convergence electronic.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(ConvergenceElectronic, self).__init__(name, parser, *args, **kwargs)

    def _serialize(self):
        return {
            'name': self.name,
            'units': self.manifest["defaults"]["units"],
            'data': self.parser.convergence_electronic()
        }



from tests.fixtures.nwchem.references import *
from tests.integration import IntegrationTestBase
from express.parsers.apps.nwchem.parser import NwchemParser


class TestNwchemParser(IntegrationTestBase):
    def setUp(self):
        super(TestNwchemParser, self).setUp()
        self.parser = NwchemParser(work_dir=self.workDir, stdout_file=self.stdoutFile)

    def tearDown(self):
        super(TestNwchemParser, self).setUp()

    def test_nwchem_total_energy(self):
        self.assertAlmostEqual(self.parser.total_energy(), TOTAL_ENERGY, places=2)

    def test_nwchem_total_energy_contributions(self):
        self.assertDeepAlmostEqual(self.parser.total_energy_contributions(), TOTAL_ENERGY_CONTRIBUTION, places=2)



from tests.unit import UnitTestBase
from tests.fixtures.data import PHONON_DISPERSIONS_RAW_DATA
from express.properties.non_scalar.two_dimensional_plot.phonon_dispersions import PhononDispersions

PHONON_DISPERSIONS = {
    "yDataSeries": [
        [
            -6e-06,
            -6.859784
        ],
        [
            -6e-06,
            -6.859784
        ]
    ],
    "xDataArray": [
        [
            0.0,
            0.0,
            0.0
        ],
        [
            0.0,
            0.05,
            0.05
        ]
    ],
    "name": "phonon_dispersions",
    "xAxis": {
        "units": "crystal",
        "label": "qpoints"
    },
    "yAxis": {
        "units": "cm-1",
        "label": "frequency"
    }
}


class PhononDispersionsTest(UnitTestBase):
    def setUp(self):
        super(PhononDispersionsTest, self).setUp()

    def tearDown(self):
        super(PhononDispersionsTest, self).setUp()

    def test_phonon_dispersions(self):
        parser = self.get_mocked_parser("phonon_dispersions", PHONON_DISPERSIONS_RAW_DATA)
        property_ = PhononDispersions("phonon_dispersions", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), PHONON_DISPERSIONS)



from tests.unit import UnitTestBase
from express.properties.scalar.p_norm import PNorm
from tests.fixtures.data import ELEMENTAL_RATIOS_RAW_DATA

P_NORM = {
    "degree": 0,
    "name": "p-norm",
    "value": 2
}


class PNormTest(UnitTestBase):
    def setUp(self):
        super(PNormTest, self).setUp()

    def tearDown(self):
        super(PNormTest, self).setUp()

    def test_p_norm(self):
        parser = self.get_mocked_parser("elemental_ratios", ELEMENTAL_RATIOS_RAW_DATA)
        property_ = PNorm("p-norm", parser, degree=0)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), P_NORM)



from express.properties.non_scalar import NonScalarProperty


class StressTensor(NonScalarProperty):
    """
    3x3 matrix expressing stresses in x, y and z dimensions.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(StressTensor, self).__init__(name, parser, *args, **kwargs)

    def _serialize(self):
        return {
            'name': self.name,
            "units": self.manifest["defaults"]["units"],
            "value": self.parser.stress_tensor()
        }



from express.properties.non_scalar.two_dimensional_plot import TwoDimensionalPlotProperty


class PotentialProfile(TwoDimensionalPlotProperty):
    """
    Potential profile.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(PotentialProfile, self).__init__(name, parser, *args, **kwargs)
        potential_profile = self.parser.potential_profile()
        self.xDataArray = potential_profile[0]
        self.yDataSeries = [potential_profile[1], potential_profile[2], potential_profile[3]]



import io
import pymatgen as mg
from ase.io import read, write

from express.parsers import BaseParser
from express.parsers.mixins.ionic import IonicDataMixin

STRUCTURE_MAP = {
    "primitive": lambda s: mg.symmetry.analyzer.SpacegroupAnalyzer(s).get_primitive_standard_structure(),
    "conventional": lambda s: mg.symmetry.analyzer.SpacegroupAnalyzer(s).get_conventional_standard_structure()
}


class StructureParser(BaseParser, IonicDataMixin):
    """
    Structure parser class.

    Args:
        args (list): args passed to the parser.
        kwargs (dict): kwargs passed to the parser.
            structure_string (str): structure string.
            structure_format (str): structure format, poscar or espresso-in.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.structure_string = kwargs.get("structure_string")
        self.structure_format = kwargs.get("structure_format")

        # convert espresso input into poscar
        if self.structure_format == "espresso-in":
            self.structure_format = "poscar"
            self.structure_string = self.espresso_input_to_poscar(self.structure_string)

        # cell_type is either original, primitive or conventional
        self.cell_type = kwargs["cell_type"]
        self.structure = mg.Structure.from_str(self.structure_string, self.structure_format)
        if self.cell_type != "original": self.structure = STRUCTURE_MAP[self.cell_type](self.structure)

        # keep only one atom inside the basis in order to have the original lattice type
        self.lattice_only_structure = mg.Structure.from_str(self.structure_string, self.structure_format)  # deepcopy
        self.lattice_only_structure.remove_sites(range(1, len(self.structure.sites)))


    def lattice_vectors(self):
        """
        Returns lattice vectors.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.lattice_vectors
        """
        return {
            'vectors': {
                'a': self.structure.lattice.matrix.tolist()[0],
                'b': self.structure.lattice.matrix.tolist()[1],
                'c': self.structure.lattice.matrix.tolist()[2],
                'alat': 1.0
            }
        }

    def lattice_bravais(self):
        """
        Returns lattice bravais.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.lattice_bravais
        """
        return {
            "type": self._lattice_type(),
            "a": self.structure.lattice.a,
            "b": self.structure.lattice.b,
            "c": self.structure.lattice.c,
            "alpha": self.structure.lattice.alpha,
            "beta": self.structure.lattice.beta,
            "gamma": self.structure.lattice.gamma,
            "units": {
                "length": "angstrom",
                "angle": "degree"
            }
        }

    def _lattice_type(self):
        """
        Returns lattice type according to AFLOW (http://aflowlib.org/) classification.

        Returns:
             str
        """
        structure_ = self.lattice_only_structure if self.cell_type != "primitive" else self.structure
        try:
            # try getting the lattice type from the lattice only structure
            return self._lattice_type_from_structure(structure_)
        except:
            try:
                # try getting the lattice type from the current structure
                return self._lattice_type_from_structure(self.structure)
            except:
                return "TRI"

    def _lattice_type_from_structure(self, structure_):
        """
        Returns lattice type according to AFLOW (http://aflowlib.org/) classification.

        Returns:
             str
        """
        analyzer = mg.symmetry.analyzer.SpacegroupAnalyzer(structure_, symprec=0.001)
        lattice_type = analyzer.get_lattice_type()
        spg_symbol = analyzer.get_space_group_symbol()

        # TODO: find a better implementation
        if lattice_type == "cubic":
            if "P" in spg_symbol:
                return "CUB"
            elif "F" in spg_symbol:
                return "FCC"
            elif "I" in spg_symbol:
                return "BCC"
        elif lattice_type == "tetragonal":
            if "P" in spg_symbol:
                return "TET"
            elif "I" in spg_symbol:
                return "BCT"
        elif lattice_type == "orthorhombic":
            if "P" in spg_symbol:
                return "ORC"
            elif "F" in spg_symbol:
                return "ORCF"
            elif "I" in spg_symbol:
                return "ORCI"
            elif "C" in spg_symbol:
                return "ORCC"
        elif lattice_type == "hexagonal":
            return "HEX"
        elif lattice_type == "rhombohedral":
            return "RHL"
        elif lattice_type == "monoclinic":
            if "P" in spg_symbol:
                return "MCL"
            elif "C" in spg_symbol:
                return "MCLC"

        return "TRI"

    def basis(self):
        """
        Returns basis.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.basis
        """
        return {
            'units': 'crystal',
            'elements': [{'id': i + 1, 'value': v.species_string} for i, v in enumerate(self.structure.sites)],
            'coordinates': [{'id': i + 1, 'value': v.frac_coords.tolist()} for i, v in enumerate(self.structure.sites)]
        }

    def space_group_symbol(self):
        """
        Returns space group symbol.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.space_group_symbol
        """
        return {
            "value": mg.symmetry.analyzer.SpacegroupAnalyzer(self.structure).get_space_group_symbol(),
            "tolerance": 0.3
        }

    def formula(self):
        """
        Returns formula.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.formula
        """
        return self.structure.composition.formula

    def reduced_formula(self):
        """
        Returns reduced formula.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.reduced_formula
        """
        return self.structure.composition.reduced_formula

    def elemental_ratios(self):
        """
        Returns elemental ratios.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.elemental_ratios
        """
        return {
            el.symbol: self.structure.composition.get_atomic_fraction(el) for el in self.structure.composition.elements
        }

    def atomic_constraints(self):
        """
        Returns atomic constraints.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.atomic_constraints
        """
        return self.structure.site_properties.get("selective_dynamics")

    def espresso_input_to_poscar(self, espresso_input):
        """
        Extracts structure from espresso input file and returns in poscar format.

        Args:
            espresso_input (str): input file content

        Returns:
            str: poscar
        """
        input_ = io.StringIO()
        input_.write(espresso_input)
        input_.seek(0)
        atoms = read(input_, format="espresso-in")
        output_ = io.StringIO()
        write(output_, atoms, format="vasp", vasp5=True)
        content = output_.getvalue()
        input_.close()
        output_.close()
        return content



import numpy as np

from copy import deepcopy
from express import settings
from express.properties.utils import eigenvalues
from express.properties.non_scalar import NonScalarProperty


class BandGaps(NonScalarProperty):
    """
    The minimum energy difference between the highest occupied (valence) band and the lowest unoccupied ( conduction)
    band, extracted from the k-mesh of electronic eigenvalues (default) or the bandstructure. Can be direct - when
    the difference is between the energy points at the same point in reciprocal space, and indirect - when the
    difference between two inequivalent points. The flavor defines the type of band_gap computation. It could be done
    either using 'mesh' or 'path'.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(BandGaps, self).__init__(name, parser, *args, **kwargs)

        self.values = None
        self.nspins = self.safely_invoke_parser_method("nspins")
        self.ibz_k_points = self.safely_invoke_parser_method("ibz_k_points")
        self.fermi_energy = self.safely_invoke_parser_method("fermi_energy")
        self.band_gaps_direct = self.safely_invoke_parser_method("band_gaps_direct")
        self.band_gaps_indirect = self.safely_invoke_parser_method("band_gaps_indirect")
        self.eigenvalues_at_kpoints = self.safely_invoke_parser_method("eigenvalues_at_kpoints")

        if self.band_gaps_direct is not None and self.band_gaps_indirect is not None:
            self.values = [
                self._serialize_band_gaps(self.band_gaps_direct, "direct"),
                self._serialize_band_gaps(self.band_gaps_indirect, "indirect")
            ]

    def _serialize(self):
        return {
            'name': self.name,
            'values': self.values if self.values else [self.compute_on_mesh(type_) for type_ in ["direct", "indirect"]],
            'eigenvalues': self._eigenvalues() if not self.values else []
        }

    def _serialize_band_gaps(self, gap, type_):
        return {
            'type': type_,
            'units': self.manifest["defaults"]["units"],
            'value': gap
        }

    def compute_on_mesh(self, type_="indirect"):
        """
        Calculates the band gap on the material's mesh.

        Args:
            type_ (str): band gap type, either direct or indirect.

        Returns:
            dict
        """
        nk, ns, N_sk, e_skn = self._get_bands_info()
        ev_sk, ec_sk = self._get_valence_conduction_bands(e_skn)

        if ns == 1:
            gap, k1, k2 = self._find_gap(N_sk[0], ev_sk[0], ec_sk[0], type=type_)
        else:
            gap, k1, k2 = self._find_gap(N_sk.ravel(), ev_sk.ravel(), ec_sk.ravel(), type=type_)
            k1 = divmod(k1, nk)
            k2 = divmod(k2, nk)

        result = self._serialize_band_gaps(gap, type_)
        if k1 is not None and k2 is not None:
            result.update({
                'kpointValence': self._round(self.ibz_k_points[k1[1] if isinstance(k1, tuple) else k1]),
                'kpointConduction': self._round(self.ibz_k_points[k2[1] if isinstance(k2, tuple) else k2])
            })
        return result

    def _get_bands_info(self):
        """
        Extracts bands information:
            - number of kpoints (nk)
            - number of spins (ns)
            - eigenvalue mesh (e_skn)
            - number of occupied bands (N_sk)

        Returns:
            tuple: bands information containing nk, ns, e_skn and N_sk explained above.

        """
        nk = len(self.ibz_k_points)
        ns = self.nspins
        e_skn = np.array([[eigenvalues(self.eigenvalues_at_kpoints, k, s) for k in range(nk)] for s in range(ns)])
        e_skn -= self.fermi_energy
        N_sk = (e_skn < 0.0).sum(2)
        e_skn = np.array([[e_skn[s, k, N_sk[s, k] - 1:N_sk[s, k] + 1] for k in range(nk)] for s in range(ns)])
        return nk, ns, N_sk, e_skn

    def _get_valence_conduction_bands(self, e_skn):
        """
        Extracts valence and conduction bands.

        Returns:
            tuple: ev_sk (valence) and ec_sk (conduction) bands.
        """
        ev_sk = e_skn[:, :, 0]  # valence band
        ec_sk = e_skn[:, :, 1]  # conduction band
        return ev_sk, ec_sk

    def _find_gap(self, N_k, ev_k, ec_k, type="indirect"):
        """
        Computes the difference in energy between the highest valence band and the lowest conduction band.

        Args:
            N_k (ndarray): numbers of occupied bands.
            ev_k (ndarray): valence band.
            ec_k (ndarray): conduction band.
            type (str): band gap type, either direct or indirect.

        Returns:
            tuple: a (gap, k1, k2) tuple where k1 and k2 are the indices of the valence and conduction k-points.
        """
        if N_k.ptp() > 0:
            # Some band must be crossing fermi-level. Hence we return zero for band gap and the actual k-points
            kv = kc = np.argmax(N_k)
            return 0.0, kv, kc
        if type == "direct":
            gap_k = ec_k - ev_k
            k = gap_k.argmin()
            return gap_k[k], k, k
        kv = ev_k.argmax()
        kc = ec_k.argmin()
        return ec_k[kc] - ev_k[kv], kv, kc

    def _eigenvalues(self):
        """
        Extracts eigenvalues between last value in occupation 1 and first value in occupation 0.

        Returns:
             dict
        """
        eigens_at_kpoints = deepcopy(self.eigenvalues_at_kpoints)
        for eigens_at_kpoint in eigens_at_kpoints:
            eigens_at_kpoint["kpoint"] = self._round(eigens_at_kpoint["kpoint"])
            for eigens_at_spin in eigens_at_kpoint["eigenvalues"]:
                eigens_at_spin["energies"] = self._round(eigens_at_spin["energies"])
                eigens_at_spin["occupations"] = self._round(eigens_at_spin["occupations"])
                # occupations are empty in case of QE GW, hence sending all values.
                if len(eigens_at_spin["occupations"]) == 0: continue
                start = max(0, len(eigens_at_spin["occupations"]) - eigens_at_spin["occupations"][::-1].index(1.0) - 2)
                end = min(len(eigens_at_spin["occupations"]), eigens_at_spin["occupations"].index(0.0) + 2)
                eigens_at_spin["energies"] = eigens_at_spin["energies"][start:end]
                eigens_at_spin["occupations"] = eigens_at_spin["occupations"][start:end]
        return eigens_at_kpoints

    def _round(self, array):
        return np.round(array, settings.PRECISION).tolist()



from express.parsers.settings import Constant
from express.parsers.apps.nwchem import settings
from express.parsers.formats.txt import BaseTXTParser


class NwchemTXTParser(BaseTXTParser):
    """
    Nwchem text parser class.
    """

    def __init__(self, work_dir):
        super(NwchemTXTParser, self).__init__(work_dir)

    def total_energy(self, text):
        """
        Extracts total energy.

        Args:
            text (str): text to extract data from.

        Returns:
             float
        """
        return self._general_output_parser(text, **settings.REGEX["total_energy"])

    def total_energy_contributions(self, text):
        """
        Extracts total energy contributions.

        Args:
            text (str): text to extract data from.

        Returns:
            dict
        """
        energy_contributions = {}
        for contribution in settings.TOTAL_ENERGY_CONTRIBUTIONS:
            value = self._general_output_parser(text, **settings.TOTAL_ENERGY_CONTRIBUTIONS[contribution])
            if value is not None:
                energy_contributions.update({contribution: {
                    'name': contribution,
                    'value': value
                }})
        return energy_contributions




import os

from tests.fixtures.data import SI
from tests.integration import IntegrationTestBase
from express.parsers.structure import StructureParser


class TestStructureParser(IntegrationTestBase):

    def setUp(self):
        super(IntegrationTestBase, self).setUp()

    def tearDown(self):
        super(IntegrationTestBase, self).setUp()

    @property
    def parser(self):
        manifest = self.getManifest()
        with open(os.path.join(self.rootDir, manifest["structurePath"])) as f:
            kwargs = {
                "structure_string": f.read(),
                "cell_type": manifest.get("cell_type", "original"),
                "structure_format": manifest.get("structureFormat", "poscar")
            }
            return StructureParser(**kwargs)

    def test_structure_espresso_basis(self):
        self.assertDeepAlmostEqual(self.parser.basis(), SI["basis"], places=2)

    def test_structure_vasp_basis(self):
        self.assertDeepAlmostEqual(self.parser.basis(), SI["basis"], places=2)

    def test_structure_espresso_lattice_bravais(self):
        self.assertDeepAlmostEqual(self.parser.lattice_bravais(), SI["lattice"], places=2)

    def test_structure_vasp_lattice_bravais(self):
        self.assertDeepAlmostEqual(self.parser.lattice_bravais(), SI["lattice"], places=2)



import os
from typing import Dict, Any

from . import NonScalarProperty


class FileContent(NonScalarProperty):
    """
    file_content property.
    Args:
        name (str): Name of the property
        parser (Any): Express parser being used
        basename (str): the file name, without the path
        filetype (str): text or image
        upload_dir (str): Path Express sees to the working directory where the job file exists
        object_storage_data (dict): describes the object in the bucket, with keys:
            - CONTAINER (str): the name of the object store bucket
            - NAME (str): the object name in the container
            - PROVIDER (str): the cloud provider where the container exists
            - REGION (str): the cloud provider region in which the container exists
    """

    def __init__(self, name: str, parser: Any,
                 *args,
                 basename: str, filetype: str, upload_dir: str, object_storage_data: Dict[str, str],
                 **kwargs):
        super().__init__(name, parser, *args, **kwargs)
        self.basename = basename
        self.filetype = filetype
        self.upload_dir = upload_dir
        self.object_storage_data: Dict[str, str] = object_storage_data

    def _serialize(self) -> Dict:
        return {
            "name": "file_content",
            "basename": self.basename,
            "filetype": self.filetype,
            "objectData": {
                "CONTAINER": self.object_storage_data['CONTAINER'],
                "NAME": os.path.join(self.upload_dir, self.basename),
                "PROVIDER": self.object_storage_data['PROVIDER'],
                "REGION": self.object_storage_data['REGION'],
            }
        }



import mock
import os

from tests.unit import UnitTestBase
from express.properties.workflow import PyMLTrainAndPredictWorkflow
from tests.fixtures.pyML.data import WORKFLOW_TRAIN, WORKFLOW_PREDICT, NAME, PARSER, ARGS, WORK_DIR, UPLOAD_DIR
from tests.fixtures.pyML.data import CONTEXT_DIR_RELATIVE_PATH, OBJECT_STORAGE_DATA, MOCK_BASENAMES
import json


class WorkflowTest(UnitTestBase):
    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().setUp()

    @mock.patch('express.properties.workflow.os')
    def test_pyml_workflow(self, mock_os):
        mock_os.listdir.return_value = MOCK_BASENAMES

        name = NAME
        parser = PARSER
        args = ARGS
        kwargs = {
            "work_dir": WORK_DIR,
            "upload_dir": UPLOAD_DIR,
            "object_storage_data": OBJECT_STORAGE_DATA,
            "context_dir_relative_path": CONTEXT_DIR_RELATIVE_PATH,
            "workflow": WORKFLOW_TRAIN
        }

        property_ = PyMLTrainAndPredictWorkflow(name, parser, *args, **kwargs)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), WORKFLOW_PREDICT)



from express.properties.scalar import ScalarProperty


class Volume(ScalarProperty):
    """
    Volume property class.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(Volume, self).__init__(name, parser, *args, **kwargs)
        self.value = self.parser.volume()



from tests.unit import UnitTestBase
from tests.fixtures.data import MAGNETIC_MOMENTS_RAW_DATA
from express.properties.non_scalar.magnetic_moments import MagneticMoments

MAGNETIC_MOMENTS = {
    "units": "uB",
    "name": "magnetic_moments",
    "values": [
        {
            "id": 1,
            "value": [
                0,
                0,
                1.235
            ]
        },
        {
            "id": 2,
            "value": [
                0,
                0,
                -1.235
            ]
        }
    ]
}


class MagneticMomentsTest(UnitTestBase):
    def setUp(self):
        super(MagneticMomentsTest, self).setUp()

    def tearDown(self):
        super(MagneticMomentsTest, self).setUp()

    def test_atomic_forces(self):
        parser = self.get_mocked_parser("magnetic_moments", MAGNETIC_MOMENTS_RAW_DATA)
        property_ = MagneticMoments("magnetic_moments", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), MAGNETIC_MOMENTS)



from express.properties.scalar import ScalarProperty


class TotalEnergy(ScalarProperty):
    """
    The ground state energy of the system.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(TotalEnergy, self).__init__(name, parser, *args, **kwargs)
        self.value = self.parser.total_energy()



from tests.unit import UnitTestBase
from tests.fixtures.data import DOS_RAW_DATA
from express.properties.non_scalar.two_dimensional_plot.density_of_states import DensityOfStates

DOS = {
    "legend": [
        {},
        {
            "electronicState": "2py",
            "element": "Si"
        },
        {
            "electronicState": "2px",
            "element": "Si"
        },
        {
            "electronicState": "1s",
            "element": "Si"
        },
        {
            "electronicState": "2pz",
            "element": "Si"
        }
    ],
    "name": "density_of_states",
    "xAxis": {
        "label": "energy",
        "units": "eV"
    },
    "xDataArray": [
        [
            -6.005000114440918,
            -5.954999923706055,
            -5.90500020980835
        ]
    ],
    "yAxis": {
        "label": "density of states",
        "units": "states/unitcell"
    },
    "yDataSeries": [
        [
            0.00012799999967683107,
            0.0010100000072270632,
            0.006130000110715628
        ],
        [
            1.6499999980444308E-17,
            1.3080000562020133E-16,
            7.899999954541818E-16
        ]
    ]
}


class DensityOfStatesTest(UnitTestBase):
    def setUp(self):
        super(DensityOfStatesTest, self).setUp()

    def tearDown(self):
        super(DensityOfStatesTest, self).setUp()

    def test_dos(self):
        parser = self.get_mocked_parser("dos", DOS_RAW_DATA)
        property_ = DensityOfStates("density_of_states", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), DOS)



import os
from express.parsers.settings import Constant
from express.parsers import BaseParser
from express.parsers.mixins.ionic import IonicDataMixin
from express.parsers.mixins.reciprocal import ReciprocalDataMixin
from express.parsers.mixins.electronic import ElectronicDataMixin
from express.parsers.apps.nwchem.formats.txt import NwchemTXTParser
from express.parsers.apps.nwchem import settings


class NwchemParser(BaseParser, IonicDataMixin, ElectronicDataMixin, ReciprocalDataMixin):
    """
    Nwchem parser class.
    """

    def __init__(self, *args, **kwargs):
        super(NwchemParser, self).__init__(*args, **kwargs)
        self.work_dir = self.kwargs["work_dir"]
        self.stdout_file = self.kwargs["stdout_file"]
        self.txt_parser = NwchemTXTParser(self.work_dir)

    def total_energy(self):
        """
        Returns total energy.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.total_energy
            NWChem energies are defaulted to hartrees and are converted to eV in this method
        """
        total_dft_energy = Constant.HARTREE * self.txt_parser.total_energy(self._get_file_content(self.stdout_file))
        return total_dft_energy

    def total_energy_contributions(self):
        """
        Extracts total energy contributions.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.total_energy_contributions
            NWChem energies are defaulted to hartrees and are converted to eV in this method.
        """
        energy_contributions = self.txt_parser.total_energy_contributions(self._get_file_content(self.stdout_file))
        for key1, value1 in energy_contributions.items():
            for key2, value2 in value1.items():
                if type(value2) == float:
                    value1[key2] = value2 * Constant.HARTREE
        return energy_contributions

    def _is_nwchem_output_file(self, path):
        """
        Checks whether the given file is nwchem output file.
        The file is considered nwchem output if it says 'Northwest Computational Chemistry Package' at the top.

        NOTE: DO NOT READ THE WHOLE FILE INTO MEMORY AS IT COULD BE BIG.

        Returns:
             bool
        """
        if os.path.exists(path):
            with open(path, "r") as f:
                for index, line in enumerate(f):
                    if index > 25:
                        break
                    if settings.NWCHEM_OUTPUT_FILE_REGEX in line:
                        return True
                return False

    def _find_nwchem_output_files(self):
        """
        Identifies the nwchem output files for parsing.

        Returns:
            str
        """
        nwchem_output_files = []
        for root, dirs, files in os.walk(self.work_dir, followlinks=True):
            for file in files:
                path = os.path.join(root, file)
                if self._is_nwchem_output_file(path):
                    nwchem_output_files.append(path)
        return nwchem_output_files



from express.properties import BaseProperty


class NonScalarProperty(BaseProperty):
    """
    Base non-scalar property class.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(NonScalarProperty, self).__init__(name, parser, *args, **kwargs)



from express.parsers import BaseParser
from express.parsers.mixins.exabyteml import ExabyteMLDataMixin


class ExabyteMLParser(BaseParser, ExabyteMLDataMixin):
    """
    Exabyte ML parser class.
    """

    def __init__(self, *args, **kwargs):
        super(ExabyteMLParser, self).__init__(*args, **kwargs)
        self.model = kwargs.get("model")
        self.targets = kwargs.get("targets")
        self.features = kwargs.get("features")
        self.predicted_properties = self.kwargs.get("predicted_properties")
        self.scaling_params_per_feature = kwargs.get("scaling_params_per_feature")

    def band_gaps_direct(self):
        return next((i["value"] for i in self.predicted_properties if i["name"] == "band_gaps:direct"))

    def band_gaps_indirect(self):
        return next((i["value"] for i in self.predicted_properties if i["name"] == "band_gaps:indirect"))



from tests.unit import UnitTestBase
from tests.fixtures.data import STRESS_TENSOR_RAW_DATA
from express.properties.non_scalar.stress_tensor import StressTensor

STRESS_TENSOR = {
    "name": "stress_tensor",
    "value": [
        [
            3,
            0,
            0
        ],
        [
            0,
            3,
            0
        ],
        [
            0,
            0,
            3
        ]
    ],
    "units": "kbar"
}


class StressTensorTest(UnitTestBase):
    def setUp(self):
        super(StressTensorTest, self).setUp()

    def tearDown(self):
        super(StressTensorTest, self).setUp()

    def test_stress_tensor(self):
        parser = self.get_mocked_parser("stress_tensor", STRESS_TENSOR_RAW_DATA)
        property_ = StressTensor("stress_tensor", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), STRESS_TENSOR)



from express.properties.scalar import ScalarProperty


class SurfaceEnergy(ScalarProperty):

    def __init__(self, name, parser, *args, **kwargs):
        super(SurfaceEnergy, self).__init__(name, parser, *args, **kwargs)
        self.value = kwargs["value"]



from abc import abstractmethod


class IonicDataMixin(object):
    """
    Defines ionic interfaces.

    Note:
        THE FORMAT OF DATA STRUCTURE RETURNED MUST BE PRESERVED IN IMPLEMENTATION.
    """

    @abstractmethod
    def initial_lattice_vectors(self):
        """
        Returns initial lattice vectors.

        Units: angstrom

        Returns:
            dict

        Example:
            {
                'a': [-0.561154473, -0.000000000, 0.561154473],
                'b': [-0.000000000, 0.561154473, 0.561154473],
                'c': [-0.561154473, 0.561154473, 0.000000000],
                'alat': 9.44858082
             }
        """
        pass

    @abstractmethod
    def initial_basis(self):
        """
        Returns initial basis.

        Units: crystal

        Returns:
            dict

        Example:
            {
                'units': 'crystal',
                'elements': [{'id': 1, 'value': 'Si'}, {'id': 2, 'value': 'Si'}],
                'coordinates': [{'id': 1, 'value': [0.0, 0.0, 0.0]}, {'id': 2, 'value': [0.25, 0.25, 0.25]}]
             }
        """
        pass

    @abstractmethod
    def final_lattice_vectors(self):
        """
        Returns final lattice vectors.

        Units: angstrom

        Returns:
            dict

        Example:
            {
                'a': [-0.561154473, -0.000000000, 0.561154473],
                'b': [-0.000000000, 0.561154473, 0.561154473],
                'c': [-0.561154473, 0.561154473, 0.000000000],
                'alat': 9.44858082
             }
        """
        pass

    @abstractmethod
    def final_basis(self):
        """
        Returns final basis.

        Units: crystal

        Returns:
            dict

        Example:
            {
                'units': 'crystal',
                'elements': [{'id': 1, 'value': 'Si'}, {'id': 2, 'value': 'Si'}],
                'coordinates': [{'id': 1, 'value': [0.0, 0.0, 0.0]}, {'id': 2, 'value': [0.25, 0.25, 0.25]}]
             }
        """
        pass

    @abstractmethod
    def convergence_ionic(self):
        """
        Extracts convergence ionic.

        Units:
            energy: eV

        Returns:
             list[dict]

        Example:
            [
                {
                    "electronic": {
                        "data": [
                            0.10304755,
                            0.04365706,
                            0.00051533,
                            0.00007517,
                            0.00000294,
                            6.7E-7
                        ],
                        "units": "eV"
                    },
                    "energy": -258.629394472941,
                    "structure": {
                        "basis": {
                            "coordinates": [
                                {
                                    "id": 0,
                                    "value": [
                                        0,
                                        0,
                                        0
                                    ]
                                },
                                {
                                    "id": 1,
                                    "value": [
                                        1.11631153838261,
                                        0.789351104951054,
                                        1.93350786347877
                                    ]
                                }
                            ],
                            "elements": [
                                {
                                    "id": 0,
                                    "value": "Si"
                                },
                                {
                                    "id": 1,
                                    "value": "Si"
                                }
                            ],
                            "units": "angstrom"
                        },
                        "lattice": {
                            "vectors": {
                                "a": [
                                    3.34893229493841,
                                    0,
                                    1.93350786347877
                                ],
                                "alat": 1,
                                "b": [
                                    1.11631076497947,
                                    3.15740674001365,
                                    1.93350786347877
                                ],
                                "c": [
                                    0,
                                    0,
                                    3.86701572695754
                                ]
                            }
                        }
                    }
                }
            ]
        """
        pass

    @abstractmethod
    def stress_tensor(self):
        """
        Returns stress tensor.

        Units: kbar

        Returns:
            list

        Example:
            [
                [
                    0.00050115,
                    -1e-08,
                    0.0
                ],
                [
                    -1e-08,
                    0.0005011,
                    0.0
                ],
                [
                    0.0,
                    -0.0,
                    0.00050111
                ]
            ]
        """
        pass

    @abstractmethod
    def pressure(self):
        """
        Returns pressure.

        Units: kbar

        Returns:
            float

        Examples:
             73.72
        """
        pass

    @abstractmethod
    def total_force(self):
        """
        Returns total force.

        Units: eV/angstrom

        Returns:
            float

        Example:
            1e-06
        """
        pass

    @abstractmethod
    def atomic_forces(self):
        """
        Returns forces that is exerted on each atom by its surroundings.

        Units: eV/angstrom

        Returns:
            list

        Example:
            [
                [
                    -3.9e-07,
                    -2.4e-07,
                    0.0
                ],
                [
                    3.9e-07,
                    2.4e-07,
                    0.0
                ]
            ]
        """
        pass

    @abstractmethod
    def atomic_constraints(self):
        """
        Returns atomic constraints.

        Returns:
            list

        Example:
            [
                [
                    True,
                    False,
                    True
                ],
                [
                    False,
                    False,
                    True
                ]
            ]
        """
        pass

    @abstractmethod
    def space_group_symbol(self):
        """
        Returns space group symbol.

        Returns:
             dict

        Example:
            {
                "value": "Fd-3m",
                "tolerance": 0.3
            }
        """
        pass

    @abstractmethod
    def formula(self):
        """
        Returns formula.

        Return:
             str

        Examples:
             SiGe
        """
        pass

    @abstractmethod
    def reduced_formula(self):
        """
        Returns reduced formula.

        Return:
             str

        Examples:
             SiGe
        """
        pass

    @abstractmethod
    def volume(self):
        """
        Returns volume.

        Units: angstrom^3

        Returns:
             float

        Examples:
             1
        """
        pass

    @abstractmethod
    def elemental_ratios(self):
        """
        Returns elemental ratio.

        Returns:
             dict

        Examples:
            {
                "Si": 0.6,
                "Ge": 0.4
            }
        """
        pass

    @abstractmethod
    def density(self):
        """
        Returns density.

        Units: g/cm^3

        Returns:
             float

        Examples:
             1
        """
        pass

    @abstractmethod
    def zero_point_energy(self):
        """
        Returns zero point energy.

        Units: eV

        Returns:
             float

        Examples:
             73.72
        """
        pass

    @abstractmethod
    def phonon_dos(self):
        """
        Returns phonon dos.

        Returns:
             dict

        Example:
            {
                'frequency': [-1.2588E-05, 9.9999E-01, 2.0000E+00, 3.0000E+00, ....]
                'total': [0.0000E+00, 2.5386E-07, 1.0154E-06, 2.2847E-06, ....]
            }
        """
        pass

    @abstractmethod
    def phonon_dispersions(self):
        """
        Returns phonon dispersions.

        Returns:
             dict

        Example:
            {
                'qpoints': [[0.00, 0.00, 0.00],[0.00, 0.00, 0.01],....],
                'frequencies': [['-0.0000', '-0.0000', '-0.0000', '574.0778', '574.0778', '574.2923'],
                ['29.3716', '30.0630', '70.4699', '568.0790', '569.7664', '569.9710'], ....]
            }
        """
        pass

    @abstractmethod
    def magnetic_moments(self):
        """
        Returns magnetic moments.

        Units: uB

        Returns:
            list

        Examples:
             [[0, 0, 3.046], [0, 0, -3.045]]
        """
        pass



import numpy as np

from express.properties.utils import eigenvalues
from express.settings import ZERO_WEIGHT_KPOINT_THRESHOLD
from express.properties.non_scalar.two_dimensional_plot import TwoDimensionalPlotProperty


class BandStructure(TwoDimensionalPlotProperty):
    """
    Describes the ranges of energy that an electron within the solid may have (called energy bands, allowed bands,
    or simply bands) and ranges of energy that it may not have (called band gaps or forbidden bands). By having
    'user' boolean type argument one could get the user defined band structure or the default one respectively. This
    is a complex property containing energy values for a number of electronic bands on a path in a three-dimensional
    reciprocal space.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(BandStructure, self).__init__(name, parser, *args, **kwargs)
        self.nspins = self.parser.nspins()

        self.eigenvalues_at_kpoints = self.parser.eigenvalues_at_kpoints()
        if kwargs.get("remove_non_zero_weight_kpoints", False):
            self.eigenvalues_at_kpoints = [e for e in self.eigenvalues_at_kpoints if e['weight'] <= ZERO_WEIGHT_KPOINT_THRESHOLD]

        self.nkpoints = len(self.eigenvalues_at_kpoints)
        self.bands = self._get_band()
        self.xDataArray = [eigenvalueData["kpoint"] for eigenvalueData in self.eigenvalues_at_kpoints]
        self.yDataSeries = self.bands.tolist()

    def _serialize(self):
        data = super(BandStructure, self)._serialize()
        data.update({'spin': [0.5, -0.5] * len(self.bands) if self.nspins > 1 else [0.5] * len(self.bands)})
        return data

    def _get_band(self):
        """
        Returns bands.

        Returns:
            ndarray
        """
        bands = np.array([[eigenvalues(self.eigenvalues_at_kpoints, k, s) for s in range(self.nspins)] for k in range(self.nkpoints)])
        bands = np.transpose(bands).reshape(len(bands[0][0]) * self.nspins, self.nkpoints)
        return bands



import os


class BaseParser(object):
    """
    Base Parser class.
    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def _get_file_content(self, file_path):
        """
        Returns the content of a given file.

        Args:
            file_path (str): file path.

        Returns:
             str
        """
        content = ""
        if file_path and os.path.exists(file_path):
            with open(file_path) as f:
                content = f.read()
        return content



import os
import logging

from express.properties import BaseProperty
from express.properties.scalar.p_norm import PNorm
from express.properties.scalar.volume import Volume
from express.parsers.structure import StructureParser
from express.properties.scalar.density import Density
from express.parsers.apps.vasp.parser import VaspParser
from express.parsers.utils import lattice_basis_to_poscar
from express.properties.non_scalar.symmetry import Symmetry
from express.properties.scalar.elemental_ratio import ElementalRatio
from express.properties.structural.inchi import Inchi
from express.properties.structural.inchi_key import InchiKey
from express.parsers.molecule import MoleculeParser
from express.parsers.crystal import CrystalParser

class Material(BaseProperty):
    """
    Material class.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(Material, self).__init__(name, parser, *args, **kwargs)
        self.is_non_periodic = kwargs.get("is_non_periodic", False)

        cell_type = kwargs.get("cell_type", "original")
        structure_string = kwargs.get("structure_string")
        structure_format = kwargs.get("structure_format", "poscar")

        if not structure_string:
            if kwargs.get("is_initial_structure"):
                if isinstance(self.parser, VaspParser):
                    with open(os.path.join(self.parser.work_dir, "POSCAR")) as f:
                        structure_string = f.read()
                else:
                    basis = self.parser.initial_basis()
                    lattice = self.parser.initial_lattice_vectors()
                    structure_string = lattice_basis_to_poscar(lattice, basis)

            if kwargs.get("is_final_structure"):
                if isinstance(self.parser, VaspParser):
                    with open(os.path.join(self.parser.work_dir, "CONTCAR")) as f:
                        structure_string = f.read()
                else:
                    basis = self.parser.final_basis()
                    lattice = self.parser.final_lattice_vectors()
                    structure_string = lattice_basis_to_poscar(lattice, basis)

        if self.is_non_periodic == False:
            self.parser = CrystalParser(structure_string=structure_string, structure_format=structure_format, cell_type=cell_type)
        else:
            self.parser = MoleculeParser(structure_string=structure_string, structure_format=structure_format, cell_type=cell_type)

    @property
    def formula(self):
        return self.parser.reduced_formula()

    @property
    def unitCellFormula(self):
        return self.parser.formula()

    @property
    def derived_properties(self):
        derived_properties = []
        try:
            symmetry = Symmetry("symmetry", self.parser).serialize_and_validate()
            if self.is_non_periodic:
                inchi = Inchi("inchi", self.parser).serialize_and_validate()
                inchi_key = InchiKey("inchi_key", self.parser).serialize_and_validate()
                volume = None
                density = None
                derived_properties = [symmetry, inchi, inchi_key]
            else:
                inchi = None
                inchi_key = None
                volume = Volume("volume", self.parser).serialize_and_validate()
                density = Density("density", self.parser).serialize_and_validate()
                derived_properties = [volume, density, symmetry]
            derived_properties.extend(self._elemental_ratios())
            derived_properties.extend(self._p_norms())
        # TODO: Determine how to avoid an eternal pass when one derived property fails
        except:
            logging.info("Derived properties array empty due to failure to caluclate one (or more) values.")
            pass
        return derived_properties

    @property
    def basis(self):
        return self.parser.basis()

    @property
    def lattice(self):
        return self.parser.lattice_bravais()

    def _serialize(self):
        """
        Serialize a material.

        Returns:
             dict
        """
        return {
            "_id": "",
            "name": self.name,
            "exabyteId": "",
            "hash": "",
            "formula": self.formula,
            "unitCellFormula": self.unitCellFormula,
            "lattice": self.lattice,
            "basis": self.basis,
            "derivedProperties": self.derived_properties,
            "creator": {
                "_id": "",
                "cls": "User",
                "slug": ""
            },
            "owner": {
                "_id": "",
                "cls": "Account",
                "slug": ""
            },
            "schemaVersion": "0.2.0",
        }

    def _elemental_ratios(self):
        """
        Returns a list of elemental ratios.

        Returns:
             list
        """
        elemental_ratios = []
        for element in self.parser.elemental_ratios().keys():
            elemental_ratio = ElementalRatio("elemental_ratio", self.parser, element=element).serialize_and_validate()
            elemental_ratios.append(elemental_ratio)
        return elemental_ratios

    def _p_norms(self):
        """
        Returns a list of p-norms.

        Reference:
            https://images.nature.com/full/nature-assets/npjcompumats/2016/npjcompumats201628/extref/npjcompumats201628-s1.pdf

        Returns:
            list
        """
        p_norms = []
        for degree in [0, 2, 3, 5, 7, 10]:
            p_norms.append(PNorm("p-norm", self.parser, degree=degree).serialize_and_validate())
        return p_norms



import string
import numpy as np
from operator import itemgetter

from express.parsers.formats.xml import BaseXMLParser

SPIN_MAP_COLLINEAR = {
    1: 'up',
    2: 'down'
}

SPIN_MAP_NON_COLLINEAR = {
    1: 'total',
    2: 'x',
    3: 'y',
    4: 'z'
}

EXTRACT_PARTIAL_DOS_FOR_ALL_SPINS = False


class VaspXMLParser(BaseXMLParser):
    """
    Vasp XML parser class.

    Args:
        xml_file_path (str): path to the xml file.
    """

    def __init__(self, xml_file_path):
        super(VaspXMLParser, self).__init__(xml_file_path)

    def eigenvalues_at_kpoints(self):
        """
        Returns eigenvalues for all kpoints.

        Returns:
             list

        Example:
            [
                {
                    'kpoint': [-0.5, 0.5, 0.5],
                    'weight': 9.5238095E-002,
                    'eigenvalues': [
                        {
                            'energies': [-1.4498446E-001, ..., 4.6507387E-001],
                            'occupations': [1, ... , 0],
                            'spin': 0.5
                        }
                    ]
                },
                ...
            ]
        """
        kpoints_list = self.root.find('kpoints').find('.//varray[@name="kpointlist"]')
        kpoints_weight = self.root.find('kpoints').find('.//varray[@name="weights"]')
        eigenvalues, occupations = self._parse_eigenvalues_occupations()

        eigenvalues_at_kpoints = []
        for kp_id, (kpoint, weight) in enumerate(zip(kpoints_list, kpoints_weight)):
            eigenvalues_at_kpoint = {
                "kpoint": [float(x) for x in kpoint.text.split()],
                "weight": float(weight.text),
                "eigenvalues": []
            }
            for spin in eigenvalues:
                eigenvalues_at_kpoint['eigenvalues'].append({
                    'energies': eigenvalues[spin][kp_id].tolist(),
                    'occupations': occupations[spin][kp_id].tolist(),
                    'spin': 0.5 if spin == 0 else -0.5
                })
            eigenvalues_at_kpoints.append(eigenvalues_at_kpoint)

        return eigenvalues_at_kpoints

    def _parse_eigenvalues_occupations(self):
        """
        Extracts eigenvalues and occupations for each spin and kpoint.

        Returns:
            tuple: a tuple of the eigenvalues and occupations for each spin and kpoint.
        """
        eigenvalues = {}
        occupations = {}
        eigenvalues_tag = self.root.findall('calculation')[-1].find('eigenvalues/array/set')
        for id_spin, eigen_spin in enumerate(eigenvalues_tag):
            eigenvalues[id_spin] = {}
            occupations[id_spin] = {}
            for id_kpt, eigen_kpt in enumerate(eigen_spin):
                # TODO: strip out the non-numeric values (*) for all kpoints instead of replacing them with last number.
                kpt_data = [[float(x) for x in _.text.split()] for _ in eigen_kpt if "*" not in _.text]
                kpt_data.extend([kpt_data[-1] for i in range(len(eigen_kpt) - len(kpt_data))])
                # eigenvalues mayn't be sorted properly.
                kpt_data = np.array(sorted(sorted(kpt_data, key=itemgetter(0)), key=itemgetter(1), reverse=True))
                eigenvalues[id_spin][id_kpt] = kpt_data[:, 0]
                occupations[id_spin][id_kpt] = kpt_data[:, 1]
        return eigenvalues, occupations

    def fermi_energy(self):
        """
        Extracts fermi energy.

        Returns:
            float
        """
        tag = self.root.findall('calculation')[-1].find('dos/i')
        return float(tag.text)

    def nspins(self):
        """
        Extracts the number of number of spin components.

        Returns:
             int
        """
        tag = self.root.find('parameters').find('.//separator[@name="electronic spin"]').find('.//i[@name="ISPIN"]')
        return int(tag.text)

    def dos(self, combined=True):
        """
        Extracts density of states. DOS value for each atom with the same element and orbit number will be added
        together and packed in a dictionary. The result containing energy levels, total DOS and partial DOS for each
        element will be returned.

        Args:
            combined (bool): If True, PDOS value for the same elements will be added together.

        Returns:
            dict

        Example:
            {
                'energy': [-1.0, 0, 1.0],
                'total': [0.013, 0.124, 0.923],
                'partial': [
                    {
                        'element': 'C',
                        'electronicState': 's-up',
                        'value': [0.00015, 0.000187, 0.000232, 0.000287, 0.000355, 0.000437]
                    },
                    {
                        'element': 'C',
                        'electronicState': 'p-up',
                        'value': [6.87e-06, 8.5e-06, 1.0e-05, 1.3e-05, 1.63e-05, 2.01e-05]
                    },
                ]
            }
        """
        total_dos, partial_dos_values, partial_dos_infos, electronic_states = self._extract_dos()
        if combined:
            combined_pdos_values = []
            combined_pdos_infos = []
            for atom_type in set(self.atom_names()):
                for elec_state in electronic_states:
                    matched_pdos = []
                    for pdos_idx, pdos_info_item in enumerate(partial_dos_infos):
                        if pdos_info_item['electronicState'] == elec_state and atom_type == pdos_info_item['element']:
                            matched_pdos.append(partial_dos_values[pdos_idx])
                    combined_pdos_values.append(np.sum(matched_pdos, axis=0).tolist())
                    combined_pdos_infos.append({
                        'element': atom_type,
                        'electronicState': elec_state,
                        'spin': 0.5 if 'up' in elec_state else -0.5
                    })
            partial_dos_values, partial_dos_infos = combined_pdos_values, combined_pdos_infos

        # TODO: extract and return total dos for all the spins
        return {
            'energy': total_dos[0]["energy"].tolist(),
            'total': np.sum([dos["total"] for dos in total_dos], axis=0).tolist(),
            'partial': partial_dos_values,
            'partial_info': partial_dos_infos
        }

    def atom_names(self):
        """
        Parses atoms information and returns name of atoms.

        Returns:
            list: list of atom names.
        """
        return [atom.find('c').text.strip() for atom in self.root.find('atominfo/array/set').findall('rc')]

    def _extract_total_dos(self, dos_root):
        total_dos = []
        total_dos_root = dos_root.find('total/array/set')
        for index, spin in enumerate(total_dos_root):
            tdos_spin_as_float = [[float(x) for x in tdos.text.split()] for tdos in spin.findall('r')]
            tdos_spin = np.array(tdos_spin_as_float)
            total_dos.append({
                "spin": index + 1,
                "energy": tdos_spin[:, 0],
                "total": tdos_spin[:, 1]
            })
        return total_dos

    def _extract_dos(self):

        """
        Extracts density of states (total and partial) from xml output.

        Returns:
            tuple: energy levels, total dos, partial dos and electronic states values
        """
        dos_root = self.root.findall('calculation')[-1].find('dos')
        total_dos = self._extract_total_dos(dos_root)
        partial_dos_values, partial_dos_infos, electronic_states = self._partial_dos(dos_root)
        return total_dos, partial_dos_values, partial_dos_infos, electronic_states

    def _partial_dos(self, dos_root):
        """
        Parses partial DOS for each element with its orbit value. DOS value for each atom with the same element and
        orbit number will be added together and packed in a dictionary.

        Args:
            dos_root (xml.etree.ElementTree.Element): dos root Element instance of ElementTree XML class.

        Returns:
            dict: a dictionary containing partial DOS values for each element.
                Example:
                    [
                        {
                            'element': 'C',
                            'electronicState': 's-down',
                            'value': [0.00015, 0.000187, 0.000232, 0.000287, 0.000355, 0.000437]
                        },
                        {
                            'element': 'Ti',
                            'electronicState': 'p-up',
                            'value': [6.87e-06, 8.5e-06, 1.0e-05, 1.3e-05, 1.63e-05, 2.01e-05]
                        }
                    ]
        """
        # TODO: simplify the logic and break it down into multiple functions
        partial_dos_values = []
        partial_dos_infos = []
        electronic_states = set()
        if dos_root.find('partial') is not None:
            orbit_symbols = [orbit.text.strip() for orbit in dos_root.find('partial/array').findall('field')[1:]]
            partial_root = dos_root.find('partial/array/set')
            for atom_id, atom in enumerate(partial_root):
                for spin_id, spin in enumerate(atom):
                    # extract partial dos only for the first spin in case of non-collinear calculation
                    if spin_id > 0 and len(atom) == 4 and not EXTRACT_PARTIAL_DOS_FOR_ALL_SPINS: continue
                    pdos_spin = np.array([[float(x) for x in pdos.text.split()[1:]] for pdos in spin.findall('r')])
                    for column_id, column in enumerate(pdos_spin.T):
                        elec_state = orbit_symbols[column_id - 1]
                        if len(atom) == 2:
                            elec_state = '{0}-{1}'.format(orbit_symbols[column_id], SPIN_MAP_COLLINEAR[spin_id + 1])
                        elif len(atom) == 4:
                            elec_state = '{0}-{1}'.format(orbit_symbols[column_id], SPIN_MAP_NON_COLLINEAR[spin_id + 1])
                        # orbit_symbol is missed in VASP 5.4.4, hence the below
                        elec_state = "".join(("d", elec_state)) if "x2-y2" in elec_state else elec_state
                        electronic_states.add(elec_state)
                        partial_dos_values.append(column.tolist())
                        partial_dos_infos.append({
                            'element': self.atom_names()[atom_id],
                            'index': atom_id,
                            'electronicState': elec_state,
                            'spin': 0.5 if spin_id == 0 else -0.5
                        })
        return partial_dos_values, partial_dos_infos, electronic_states

    def final_lattice_vectors(self):
        """
        Extracts lattice.

        Returns:
            dict

        Example:
            {
                'vectors': {
                    'a': [-0.561154473, -0.000000000, 0.561154473],
                    'b': [-0.000000000, 0.561154473, 0.561154473],
                    'c': [-0.561154473, 0.561154473, 0.000000000],
                    'alat': 1.0
                }
             }
        """
        vectors = {}
        for idx, vector in enumerate(
                self._parse_varray(self.root.find('structure[@name="finalpos"]/crystal/varray[@name="basis"]'))):
            vectors.update({
                string.ascii_lowercase[idx]: vector.tolist()
            })
        vectors.update({'alat': 1.0, 'units': 'angstrom'})

        return {
            'vectors': vectors
        }

    def final_basis(self):
        """
        Extract basis.

        Returns:
            dict

        Example:
            {
                'units': 'angstrom',
                'elements': [{'id': 1, 'value': 'Si'}, {'id': 2, 'value': 'Si'}],
                'coordinates': [{'id': 1, 'value': [0.0, 0.0, 0.0]}, {'id': 2, 'value': [1.11, 0.78, 1.93]}]
             }
        """
        lattice = self.final_lattice_vectors()
        lattice_matrix = np.array([lattice["vectors"][key] for key in ["a", "b", "c"]], dtype=np.float64).reshape(
            (3, 3))
        elements, coordinates = [], []
        for idx, vector in enumerate(
                self._parse_varray(self.root.find('structure[@name="finalpos"]/varray[@name="positions"]'))):
            elements.append({
                'id': idx,
                'value': self.atom_names()[idx]
            })
            coordinates.append({
                'id': idx,
                'value': np.dot(vector, lattice_matrix).tolist()
            })

        return {
            'units': 'angstrom',
            'elements': elements,
            'coordinates': coordinates
        }

    def _parse_varray(self, varray):
        """
        Parses varray elements.

        Example:
            <varray name="positions" >
                <v>       0.00000000       0.69512316       0.69512316 </v>
                <v>       0.50000000       0.80487684       0.19512316 </v>
                <v>       0.00000000       0.30487684       0.30487684 </v>
                <v>       0.50000000       0.19512316       0.80487684 </v>
                <v>       0.50000000       0.50000000       0.50000000 </v>
                <v>       0.00000000       0.00000000      -0.00000000 </v>
            </varray>

        Args:
            varray: varray xml element.

        Returns:
            ndarray: a matrix containing all the values found in the varray.
        """
        return np.array([v.text.split() for v in varray.findall('v')],
                        dtype=np.float) if varray is not None else np.array([])

    def stress_tensor(self):
        """
        Extracts stress tensor.

        Returns:
            list
        """
        return self._parse_varray(self.root.findall('calculation')[-1].find('.//varray[@name="stress"]')).tolist()

    def atomic_forces(self):
        """
        Extracts atomic forces.

        Returns:
            list
        """
        return self._parse_varray(self.root.findall('calculation')[-1].find('.//varray[@name="forces"]')).tolist()



from express.properties.non_scalar import NonScalarProperty


class Symmetry(NonScalarProperty):
    """
    Symmetry property class.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(Symmetry, self).__init__(name, parser, *args, **kwargs)

    def _serialize(self):
        return {
            "spaceGroupSymbol": self.parser.space_group_symbol()["value"],
            "tolerance": {
                "value": self.parser.space_group_symbol()["tolerance"],
                "units": "angstrom"
            },
            "name": self.name
        }



from tests.unit import UnitTestBase
from tests.fixtures.data import ELEMENTAL_RATIOS_RAW_DATA
from express.properties.scalar.elemental_ratio import ElementalRatio

ELEMENTAL_RATIO = {
    "name": "elemental_ratio",
    "value": 0.4,
    "element": "Ge"
}


class ElementalRatioTest(UnitTestBase):
    def setUp(self):
        super(ElementalRatioTest, self).setUp()

    def tearDown(self):
        super(ElementalRatioTest, self).setUp()

    def test_elemental_ratio(self):
        parser = self.get_mocked_parser("elemental_ratios", ELEMENTAL_RATIOS_RAW_DATA)
        property_ = ElementalRatio("elemental_ratio", parser, element="Ge")
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), ELEMENTAL_RATIO)



from tests.unit import UnitTestBase
from express.properties.scalar.zero_point_energy import ZeroPointEnergy

ZERO_POINT_ENERGY = {
    "units": "eV",
    "name": "zero_point_energy",
    "value": 1
}


class TotalZeroPointTest(UnitTestBase):
    def setUp(self):
        super(TotalZeroPointTest, self).setUp()

    def tearDown(self):
        super(TotalZeroPointTest, self).setUp()

    def test_zero_point_energy(self):
        parser = self.get_mocked_parser("zero_point_energy", 1)
        property_ = ZeroPointEnergy("zero_point_energy", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), ZERO_POINT_ENERGY)



from mock import MagicMock

from tests.unit import UnitTestBase
from express.properties.non_scalar.file_content import FileContent
from tests.fixtures.data import FILE_CONTENT


class FileContentTest(UnitTestBase):
    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().setUp()

    def test_file_content(self):
        name = "file_content"
        parser = None
        args = []
        kwargs = {
            "basename": "my_parity_plot.png",
            "filetype": "image",
            "upload_dir": "/cluster-001-share/groups/exaorg-uy3burw0/exaorg-uy3burw0-default/job-python-python-ml-train-organization-jJ67E6JQ6DLyF9Q8z/",
            "object_storage_data": {
                "CONTAINER": "vagrant-cluster-001",
                "NAME": "",
                "PROVIDER": "aws",
                "REGION": "us-east-1"
            }
        }
        property_ = FileContent(name, parser, *args, **kwargs)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), FILE_CONTENT)
